#!/usr/bin/python
#
# Copyright (c) 2016 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import paleomix.common.fileutils as fileutils

from paleomix.node import Node


_DEFAULT_COLORS = ("#E69F00", "#56B4E9",
                   "#009E73", "#F0E442",
                   "#0072B2", "#D55E00",
                   "#CC79A7")


class WriteSampleList(Node):
    def __init__(self, config, output_file, dependencies=()):
        self._samples = config.database.samples

        Node.__init__(self,
                      description="<WriteSampleList -> %r>" % (output_file,),
                      input_files=(config.tablefile,),
                      output_files=(output_file,),
                      dependencies=dependencies)

    def _run(self, config, temp):
        output_file, = self.output_files
        samples = self._samples
        groups = set(sample["Group(3)"] for sample in samples.itervalues())
        colors = dict(zip(groups, _DEFAULT_COLORS))

        with open(fileutils.reroot_path(temp, output_file), "w") as handle:
            handle.write("Name\tGroup\tColor\n")

            for name, sample in sorted(samples.iteritems()):
                group = sample["Group(3)"]
                color = colors[group]

                handle.write("%s\t%s\t%s\n" % (name, group, color))

            handle.write("Sample\t-\t#000000\n")

    def _teardown(self, config, temp):
        destination, = self.output_files
        source = fileutils.reroot_path(temp, destination)

        fileutils.move_file(source, destination)



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from __future__ import print_function

import errno
import logging
import multiprocessing
import os
import pickle
import Queue
import signal
import traceback

import paleomix.ui
import paleomix.logger

from paleomix.node import \
    Node, \
    NodeError, \
    NodeUnhandledException
from paleomix.nodegraph import \
    FileStatusCache, \
    NodeGraph, \
    NodeGraphError
from paleomix.common.utilities import \
    safe_coerce_to_tuple, \
    fast_pickle_test
from paleomix.common.versions import \
    VersionRequirementError


class Pypeline(object):
    def __init__(self, config):
        self._nodes = []
        self._config = config
        self._logger = logging.getLogger(__name__)
        # Set if a keyboard-interrupt (SIGINT) has been caught
        self._interrupted = False
        self._queue = multiprocessing.Queue()
        self._pool = multiprocessing.Pool(1, _init_worker, (self._queue,))

    def add_nodes(self, *nodes):
        for subnodes in safe_coerce_to_tuple(nodes):
            for node in safe_coerce_to_tuple(subnodes):
                if not isinstance(node, Node):
                    raise TypeError("Node object expected, recieved %s"
                                    % repr(node))
                self._nodes.append(node)

    def run(self, max_threads=1, dry_run=False, progress_ui="verbose"):
        if max_threads < 1:
            raise ValueError("Max threads must be >= 1")
        _update_nprocesses(self._pool, max_threads)

        try:
            nodegraph = NodeGraph(self._nodes)
        except NodeGraphError, error:
            self._logger.error(error)
            return False

        for node in nodegraph.iterflat():
            if node.threads > max_threads:
                message = "Node(s) use more threads than the max allowed; " \
                          "the pipeline may therefore use more than the " \
                          "expected number of threads.\n"
                paleomix.ui.print_warn(message)
                break

        if dry_run:
            progress_printer = paleomix.ui.RunningUI()
            nodegraph.add_state_observer(progress_printer)
            progress_printer.flush()
            progress_printer.finalize()
            self._logger.info("Dry run done ...")
            return True

        old_handler = signal.signal(signal.SIGINT, self._sigint_handler)
        try:
            return self._run(nodegraph, max_threads, progress_ui)
        finally:
            signal.signal(signal.SIGINT, old_handler)

        return False

    def _run(self, nodegraph, max_threads, progress_ui):
        # Dictionary of nodes -> async-results
        running = {}
        # Set of remaining nodes to be run
        remaining = set(nodegraph.iterflat())

        is_ok = True
        progress_printer = paleomix.ui.get_ui(progress_ui)
        progress_printer.max_threads = max_threads
        nodegraph.add_state_observer(progress_printer)

        with paleomix.ui.CommandLine() as cli:
            while running or (remaining and not self._interrupted):
                is_ok &= self._poll_running_nodes(running,
                                                  nodegraph,
                                                  self._queue)

                if not self._interrupted:  # Prevent starting of new nodes
                    self._start_new_tasks(remaining, running, nodegraph,
                                          max_threads, self._pool)

                if running:
                    progress_printer.flush()

                max_threads = cli.process_key_presses(nodegraph,
                                                      max_threads,
                                                      progress_printer)
                progress_printer.max_threads = max_threads
                _update_nprocesses(self._pool, max_threads)

        self._pool.close()
        self._pool.join()

        progress_printer.flush()
        progress_printer.finalize()

        return is_ok

    def _start_new_tasks(self, remaining, running, nodegraph, max_threads,
                         pool):
        started_nodes = []
        idle_processes = max_threads \
            - sum(node.threads for (node, _) in running.itervalues())

        if not idle_processes:
            return False

        for node in remaining:
            if not running or (idle_processes >= node.threads):
                state = nodegraph.get_node_state(node)
                if state == nodegraph.RUNABLE:
                    try:
                        # The multi-processing module relies on pickling
                        fast_pickle_test(node)
                    except pickle.PicklingError, error:
                        self._logger.error("Node cannot be pickled; please "
                                           "file a bug-report:\n"
                                           "\tNode: %s\n\tError: %s"
                                           % (self, error))
                        nodegraph.set_node_state(node, nodegraph.ERROR)
                        started_nodes.append(node)
                        continue

                    key = id(node)
                    proc_args = (key, node, self._config)
                    running[key] = (node, pool.apply_async(_call_run,
                                                           args=proc_args))
                    started_nodes.append(node)

                    nodegraph.set_node_state(node, nodegraph.RUNNING)
                    idle_processes -= node.threads
                elif state in (nodegraph.DONE, nodegraph.ERROR):
                    started_nodes.append(node)
            elif idle_processes <= 0:
                break

        for node in started_nodes:
            remaining.remove(node)

    def _poll_running_nodes(self, running, nodegraph, queue):
        errors = None
        blocking = False

        while running and not errors:
            node, proc = self._get_finished_node(queue, running, blocking)
            if not node:
                if blocking:
                    break

                blocking = True
                continue

            try:
                # Re-raise exceptions from the node-process
                proc.get()
            except (KeyboardInterrupt, SystemExit):
                raise
            except Exception, errors:
                nodegraph.set_node_state(node, nodegraph.ERROR)

                message = [str(node),
                           "  Error (%r) occurred running command:"
                           % (type(errors).__name__)]

                for line in str(errors).strip().split("\n"):
                    message.append("    %s" % (line,))
                message.append("")

                self._logger.error("\n".join(message))

            if not errors:
                nodegraph.set_node_state(node, nodegraph.DONE)

        return not errors

    @property
    def nodes(self):
        return set(self._nodes)

    def walk_nodes(self, func):
        skip_nodes = set()

        def _walk_nodes(nodes):
            for node in nodes:
                if node in skip_nodes:
                    continue
                elif not func(node):
                    return False

                skip_nodes.add(node)
                if not _walk_nodes(node.dependencies):
                    return False
            return True

        _walk_nodes(self._nodes)

    def list_input_files(self):
        """Returns a set containing the absolute path of all input files
        required by the current pipeline. These do not include any file
        generated by the pipeline itself (output files).
        """
        input_files = set()
        output_files = set()

        def collect_output_files(node):
            for filename in node.input_files:
                input_files.add(os.path.abspath(filename))

            for filename in node.output_files:
                output_files.add(os.path.abspath(filename))

            return True

        self.walk_nodes(collect_output_files)

        return input_files - output_files

    def list_output_files(self):
        cache = FileStatusCache()
        nodegraph = NodeGraph(self._nodes, lambda: cache)
        output_files = {}

        def collect_output_files(node):
            state = None
            if nodegraph.is_done(node, cache):
                state = nodegraph.DONE
                if nodegraph.is_outdated(node, cache):
                    state = nodegraph.OUTDATED

            for filename in node.output_files:
                output_files[os.path.abspath(filename)] = state

            return True

        self.walk_nodes(collect_output_files)

        return output_files

    def list_required_executables(self):
        requirements = {}

        def collect_requirements(node):
            for executable in node.executables:
                if executable not in requirements:
                    requirements[executable] = set()

            for requirement in node.requirements:
                if requirement.name not in requirements:
                    requirements[requirement.name] = set()

                requirements[requirement.name].add(requirement)

                executable = requirement.executable
                if not requirements.get(executable):
                    requirements.pop(executable, None)

            return True

        self.walk_nodes(collect_requirements)
        return requirements

    def print_output_files(self, print_func=print):
        output_files = self.list_output_files()

        for filename, state in sorted(output_files.iteritems()):
            if state == NodeGraph.DONE:
                state = "Ready      "
            elif state == NodeGraph.OUTDATED:
                state = "Outdated   "
            else:
                state = "Missing    "

            print_func("%s\t%s" % (state, filename))

    def print_input_files(self, print_func=print):
        """Prints the absolute path of all input files required by the current
        pipeline, excluding any file generated by the pipeline itself (output
        files). One file is printed per line.
        """
        input_files = self.list_input_files()
        for filename in sorted(input_files):
            print_func("%s" % (filename,))

    def print_required_executables(self, print_func=print):
        template = "{: <40s} {: <11s} {}"
        pipeline_executables = self.list_required_executables()
        print_func(template.format("Executable",
                                   "Version",
                                   "Required version"))

        for (name, requirements) in sorted(pipeline_executables.items()):
            if not requirements:
                print_func(template.format(name, "-", "any version"))
                continue

            for requirement in requirements:
                try:
                    if requirement.version:
                        version = "v" + ".".join(map(str, requirement.version))
                    else:
                        version = "NA"
                except VersionRequirementError:
                    version = "UNKNOWN"

                print_func(template.format(name, version, requirement.checks))

    def _sigint_handler(self, signum, frame):
        """Signal handler; see signal.signal."""
        if not self._interrupted:
            self._interrupted = True
            self._logger.error("\nKeyboard interrupt detected, waiting for "
                               "running tasks to complete ... Press CTRL-C "
                               "again to force termination.\n")
        else:
            self._pool.terminate()
            raise signal.default_int_handler(signum, frame)

    def to_dot(self, destination):
        """Writes a simlpe dot file to the specified destination, representing
        the full dependency tree. Nodes are named by their class.
        """
        try:
            nodegraph = NodeGraph(self._nodes)
        except NodeGraphError, error:
            self._logger.error(error)
            return False

        # Dict recording all dependencies of nodes
        meta_dependencies = {}
        # Dict recording if anything depends on a speific node
        meta_rev_dependencies = {}
        for node in nodegraph.iterflat():
            selection = node.dependencies
            meta_dependencies[node] = selection
            for dep in selection:
                meta_rev_dependencies[dep] = True

        return self._write_dot(destination,
                               meta_dependencies,
                               meta_rev_dependencies)

    @classmethod
    def _write_dot(cls, destination, meta_dependencies, meta_rev_dependencies):
        """Writes simple dot file, in which each node is connected to their
        dependencies, using the object IDs as the node names. Labels are
        derived from the class names, excluding any "Node" postfix.
        """
        with open(destination, "w") as out:
            out.write("digraph G {\n")
            out.write("  graph [ dpi = 75 ];\n")
            out.write("  node [shape=record,width=.1,height=.1];\n")
            out.write("  splines=ortho;\n\n")

            for node, dependencies in meta_dependencies.iteritems():
                node_id = "Node_%i" % (id(node),)
                node_type = node.__class__.__name__
                if node_type.endswith("Node"):
                    node_type = node_type[:-4]

                rank = None
                color = "white"
                if not meta_dependencies.get(node):
                    color = "red"
                elif not meta_rev_dependencies.get(node):
                    color = "green"
                    rank = "sink"

                if rank is not None:
                    out.write("  {")
                    out.write("    rank = %s;\n  " % (rank,))

                out.write('  %s [label="%s"; fillcolor=%s; style=filled]\n'
                          % (node_id, node_type, color))

                if rank is not None:
                    out.write("  }")

                for dependency in dependencies:
                    dep_id = "Node_%i" % (id(dependency),)
                    out.write("  %s -> %s\n" % (dep_id, node_id))
                out.write("\n")

            out.write("}\n")

        return True

    @classmethod
    def _get_finished_node(cls, queue, running, blocking):
        """Returns a tuple containing a node that has finished running
        and it's async-result, or None for both if no such node could
        be found (and blocking is False), or if an interrupt occured
        while waiting for a node to finish.

        If blocking is True, the function will timeout after 0.1s.
        """
        try:
            key = queue.get(blocking, 0.1)
            return running.pop(key)
        except IOError, error:
            # User pressed ctrl-c (SIGINT), or similar event ...
            if error.errno != errno.EINTR:
                raise
        except Queue.Empty:
            pass
        return None, None


def _init_worker(queue):
    """Init function for subprocesses created by multiprocessing.Pool: Ensures
    that KeyboardInterrupts only occur in the main process, allowing us to do
    proper cleanup.
    """
    signal.signal(signal.SIGINT, signal.SIG_IGN)
    # This is a workaround to avoid having to use multiprocessing.Manager
    # to create the Queue objects; this is needed because the Manager class
    # creates it own process, which inherits the signal-handlers of the main
    # process, causing some rather odd behavior when the user causes a SIGINT.
    _call_run.queue = queue


def _call_run(key, node, config):
    """Wrapper function, required in order to call Node.run()
    in subprocesses, since it is not possible to pickle
    bound functions (e.g. self.run)"""
    try:
        return node.run(config)
    except NodeError:
        raise
    except Exception:
        message = "Unhandled error running Node:\n\n%s" \
            % (traceback.format_exc(),)

        raise NodeUnhandledException(message)
    finally:
        # See comment in _init_worker
        _call_run.queue.put(key)


def _update_nprocesses(pool, processes):
    """multiprocessing.Pool does not expose calls to change number of active
    processes, but does in fact support this for the 'maxtasksperchild' option.
    This function calls the related private functions to increase the number
    of available processes."""
    # FIXME: Catch ERRNO 11:
    # OSError: [Errno 11] Resource temporarily unavailable
    if pool._processes < processes:
        pool._processes = processes
        pool._repopulate_pool()



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os
import getpass

from paleomix.node import CommandNode
from paleomix.atomiccmd.builder import \
    AtomicJavaCmdBuilder, \
    create_customizable_cli_parameters, \
    use_customizable_cli_parameters
from paleomix.atomiccmd.sets import \
    ParallelCmds
from paleomix.common.fileutils import \
    swap_ext, \
    try_rmtree, \
    reroot_path, \
    describe_files
from paleomix.common.utilities import \
    safe_coerce_to_tuple
import paleomix.common.versions as versions
import paleomix.common.system


class PicardNode(CommandNode):
    """Base class for nodes using Picard Tools; adds an additional cleanup
    step, in order to allow the jars to be run using the same temporary folder
    as any other commands associated with the node. This is nessesary as some
    Picard tools create a large number of temporary files, leading to potential
    performance issues if these are located in the same folder.
    """

    def _teardown(self, config, temp):
        # Picard creates a folder named after the user in the temp-root
        try_rmtree(os.path.join(temp, getpass.getuser()))
        # Some JREs may create a folder for temporary performance counters
        try_rmtree(os.path.join(temp, "hsperfdata_" + getpass.getuser()))

        CommandNode._teardown(self, config, temp)


class ValidateBAMNode(PicardNode):
    def __init__(self, config, input_bam, input_index=None, output_log=None,
                 ignored_checks=(), big_genome_mode=False, dependencies=()):
        builder = picard_command(config, "ValidateSamFile")
        _set_max_open_files(builder, "MAX_OPEN_TEMP_FILES")

        if True or big_genome_mode:
            self._configure_for_big_genome(config, builder)

        builder.set_option("I", "%(IN_BAM)s", sep="=")
        for check in ignored_checks:
            builder.add_option("IGNORE", check, sep="=")

        output_log = output_log or swap_ext(input_bam, ".validated")
        builder.set_kwargs(IN_BAM=input_bam,
                           IN_INDEX=input_index,
                           OUT_STDOUT=output_log)

        description = "<Validate BAM: '%s'>" % (input_bam,)
        PicardNode.__init__(self,
                            command=builder.finalize(),
                            description=description,
                            dependencies=dependencies)

    @staticmethod
    def _configure_for_big_genome(config, builder):
        # CSI uses a different method for assigning BINs to records, which
        # Picard currently does not support.
        builder.add_option("IGNORE", "INVALID_INDEXING_BIN", sep="=")

        jar_path = os.path.join(config.jar_root, _PICARD_JAR)
        version_check = _PICARD_VERSION_CACHE[jar_path]

        try:
            if version_check.version >= (2, 19, 0):
                # Useless warning, as we do not build BAI indexes for large genomes
                builder.add_option("IGNORE", "REF_SEQ_TOO_LONG_FOR_BAI", sep="=")
        except versions.VersionRequirementError:
            pass  # Ignored here, handled elsewhere


class BuildSequenceDictNode(PicardNode):
    def __init__(self, config, reference, dependencies=()):
        self._in_reference = os.path.abspath(reference)

        builder = picard_command(config, "CreateSequenceDictionary")

        builder.set_option("R", "%(TEMP_OUT_REF)s", sep="=")
        builder.set_option("O", "%(OUT_DICT)s", sep="=")
        builder.set_kwargs(IN_REFERENCE=reference,
                           TEMP_OUT_REF=os.path.basename(reference),
                           OUT_DICT=swap_ext(reference, ".dict"))

        description = "<SequenceDictionary: '%s'>" % (reference,)

        PicardNode.__init__(self,
                            command=builder.finalize(),
                            description=description,
                            dependencies=dependencies)

    def _setup(self, _config, temp):
        # Ensure that Picard CreateSequenceDict cannot reuse any existing
        # sequence dictionaries, if the underlying files have changed.
        os.symlink(self._in_reference, reroot_path(temp, self._in_reference))


class MarkDuplicatesNode(PicardNode):
    @create_customizable_cli_parameters
    def customize(cls, config, input_bams, output_bam, output_metrics=None,
                  keep_dupes=False, dependencies=()):
        params = picard_command(config, "MarkDuplicates")
        _set_max_open_files(params, "MAX_FILE_HANDLES")

        params.set_option("OUTPUT", "%(OUT_BAM)s", sep="=")
        params.set_option("METRICS_FILE", "%(OUT_METRICS)s", sep="=")
        # Validation is mostly left to manual ValidateSamFile runs; required
        # because .csi indexed BAM records can have "invalid" bins.
        params.set_option("VALIDATION_STRINGENCY", "LENIENT", sep="=")
        params.add_multiple_options("I", input_bams, sep="=")

        if not keep_dupes:
            # Remove duplicates from output by default to save disk-space
            params.set_option("REMOVE_DUPLICATES", "True",
                              sep="=", fixed=False)

        output_metrics = output_metrics or swap_ext(output_bam, ".metrics")
        params.set_kwargs(OUT_BAM=output_bam,
                          OUT_METRICS=output_metrics)

        return {"command": params,
                "dependencies": dependencies}

    @use_customizable_cli_parameters
    def __init__(self, parameters):
        description = "<MarkDuplicates: %s>" \
            % (describe_files(parameters.input_bams),)
        PicardNode.__init__(self,
                            command=parameters.command.finalize(),
                            description=description,
                            dependencies=parameters.dependencies)


class MergeSamFilesNode(PicardNode):
    def __init__(self, config, input_bams, output_bam, dependencies=()):
        builder = picard_command(config, "MergeSamFiles")
        builder.set_option("OUTPUT", "%(OUT_BAM)s", sep="=")
        builder.set_option("SO", "coordinate", sep="=")
        # Validation is mostly left to manual ValidateSamFile runs; required
        # because .csi indexed BAM records can have "invalid" bins.
        builder.set_option("VALIDATION_STRINGENCY", "LENIENT", sep="=")
        builder.add_multiple_options("I", input_bams, sep="=")

        builder.set_kwargs(OUT_BAM=output_bam)
        description = "<Merge BAMs: %i file(s) -> '%s'>" \
            % (len(input_bams), output_bam)
        PicardNode.__init__(self,
                            command=builder.finalize(),
                            description=description,
                            dependencies=dependencies)


class MultiBAMInputNode(CommandNode):
    PIPE_FILE = "input.bam"

    def __init__(self, config, input_bams, command, index_format=None,
                 description=None, threads=1, dependencies=()):
        self._input_bams = safe_coerce_to_tuple(input_bams)
        self._index_format = index_format

        if not self._input_bams:
            raise ValueError("No input BAM files specified!")
        elif len(self._input_bams) > 1 and index_format:
            raise ValueError("BAM index cannot be required for > 1 file")
        elif index_format not in (None, ".bai", ".csi"):
            raise ValueError("Unknown index format %r" % (index_format,))

        if len(self._input_bams) > 1:
            merge = picard_command(config, "MergeSamFiles")
            merge.set_option("SO", "coordinate", sep="=")
            merge.set_option("COMPRESSION_LEVEL", 0, sep="=")
            merge.set_option("OUTPUT", "%(TEMP_OUT_BAM)s", sep="=")
            # Validation is mostly left to manual ValidateSamFile runs; this
            # is because .csi indexed BAM records can have "invalid" bins.
            merge.set_option("VALIDATION_STRINGENCY", "LENIENT", sep="=")
            merge.add_multiple_options("I", input_bams, sep="=")

            merge.set_kwargs(TEMP_OUT_BAM=self.PIPE_FILE)

            command = ParallelCmds([merge.finalize(), command])

        CommandNode.__init__(self,
                             command=command,
                             description=description,
                             threads=threads,
                             dependencies=dependencies)

    def _setup(self, config, temp):
        CommandNode._setup(self, config, temp)

        pipe_fname = os.path.join(temp, self.PIPE_FILE)
        if len(self._input_bams) > 1:
            os.mkfifo(pipe_fname)
        else:
            source_fname = os.path.abspath(self._input_bams[0])
            os.symlink(source_fname, pipe_fname)

            if self._index_format:
                os.symlink(swap_ext(source_fname, self._index_format),
                           swap_ext(pipe_fname, self._index_format))

    def _teardown(self, config, temp):
        os.remove(os.path.join(temp, self.PIPE_FILE))
        if self._index_format:
            os.remove(os.path.join(temp, swap_ext(self.PIPE_FILE,
                                                  self._index_format)))

        CommandNode._teardown(self, config, temp)


###############################################################################

_PICARD_JAR = "picard.jar"
_PICARD_VERSION_CACHE = {}


def picard_command(config, command):
    """Returns basic AtomicJavaCmdBuilder for Picard tools commands."""
    jar_path = os.path.join(config.jar_root, _PICARD_JAR)

    if jar_path not in _PICARD_VERSION_CACHE:
        params = AtomicJavaCmdBuilder(jar_path,
                                      temp_root=config.temp_root,
                                      jre_options=config.jre_options)

        # Arbitrary command, since just '--version' does not work
        params.set_option("MarkDuplicates")
        params.set_option("--version")

        requirement = versions.Requirement(call=params.finalized_call,
                                           name="Picard tools",
                                           search=r"\b(\d+)\.(\d+)\.\d+",
                                           checks=versions.GE(1, 137))
        _PICARD_VERSION_CACHE[jar_path] = requirement

    version = _PICARD_VERSION_CACHE[jar_path]
    params = AtomicJavaCmdBuilder(jar_path,
                                  temp_root=config.temp_root,
                                  jre_options=config.jre_options,
                                  CHECK_JAR=version,
                                  set_cwd=True)
    params.set_option(command)

    return params


# Fraction of per-process max open files to use
_FRAC_MAX_OPEN_FILES = 0.95
# Default maximum number of open temporary files used by Picard
_DEFAULT_MAX_OPEN_FILES = 8000


def _set_max_open_files(params, key):
    """Sets the maximum number of open files a picard process
    should use, at most. Conservatively lowered than the actual
    ulimit.
    """
    max_open_files = paleomix.common.system.get_max_open_files()
    if max_open_files:
        max_open_files = int(max_open_files * _FRAC_MAX_OPEN_FILES)

        if max_open_files < _DEFAULT_MAX_OPEN_FILES:
            params.set_option(key, max_open_files, sep="=")



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os
import copy
import StringIO

import nose.tools
from nose.tools import \
     assert_equal, \
     assert_raises
from flexmock import \
     flexmock

from paleomix.common.formats.fasta import FASTA
from paleomix.common.formats.msa import \
     MSA, \
     MSAError, \
     FASTAError


def test_dir():
    return os.path.dirname(os.path.dirname(os.path.dirname(__file__)))


def test_file(*args):
    return os.path.join(test_dir(), "data", *args)


###############################################################################
###############################################################################
# Tests for constructor

def test_msa_constructor__calls_validate():
    _mock = flexmock(MSA).should_receive('validate').at_least.once
    MSA([FASTA("NA", None, "ACGT")])


def test_msa_constructor__duplicate_names():
    records = [FASTA("Foo", None, "ACGT"),
               FASTA("Foo", None, "GTCA")]
    assert_raises(MSAError, MSA, records)


def test_msa_constructor__empty_msa():
    assert_raises(MSAError, MSA, [])


###############################################################################
###############################################################################
# Tests for 'seqlen' / len

def test_msa__len__corresponds_to_sequence_number_of_records():
    msa = MSA((FASTA("seq1",    None, "ACGCGTATGCATGCCGA"),
               FASTA("seq2",    None, "TGAACACACAGTAGGAT")))
    assert_equal(len(msa), 2)


def test_msa__seqlen__corresponds_to_sequence_lengths():
    msa = MSA((FASTA("seq1",    None, "ACGCGTATGCATGCCGA"),
               FASTA("seq2",    None, "TGAACACACAGTAGGAT")))
    assert_equal(msa.seqlen(), 17)


###############################################################################
###############################################################################
# Tests for 'exclude'

def test_msa_exclude__remove_one():
    fa_1 = FASTA("A", None, "ACGT")
    fa_2 = FASTA("B", None, "GCTA")
    initial = MSA([fa_1, fa_2])
    expected = MSA([fa_1])
    result = initial.exclude(["B"])
    assert_equal(result, expected)


def test_msa_exclude__missing_keys():
    msa = MSA([FASTA("Foo", None, "ACGT")])
    assert_raises(KeyError, msa.exclude, ["Bar"])


def test_msa_exclude__no_keys():
    msa = MSA([FASTA("Foo", None, "ACGT")])
    assert_raises(ValueError, msa.exclude, [])


###############################################################################
###############################################################################
# Tests for 'select'

def test_msa_select__remove_one():
    fa_1 = FASTA("A", None, "ACGT")
    fa_2 = FASTA("B", None, "GCTA")
    initial = MSA([fa_1, fa_2])
    expected = MSA([fa_1])
    result  = initial.select(["A"])
    assert_equal(result, expected)


def test_msa_select__missing_keys():
    msa = MSA([FASTA("Foo", None, "ACGT")])
    assert_raises(KeyError, msa.select, ["Bar"])


def test_msa_select__no_keys():
    msa = MSA([FASTA("Foo", None, "ACGT")])
    assert_raises(ValueError, msa.select, [])


###############################################################################
###############################################################################
# Tests for 'reduce'

def test_msa_reduce__no_empty_columns__no_columns_are_removed():
    fa_1 = FASTA("Name_A", "Meta_A", "ACnT")
    fa_2 = FASTA("Name_B", "Meta_B", "C-TN")
    initial = MSA([fa_1, fa_2])
    expected = MSA([fa_1, fa_2])
    assert_equal(initial.reduce(), expected)


def test_msa_reduce__one_empty_column__column_are_removed():
    fa_1 = FASTA("Name_A", "Meta_A", "AnT")
    fa_2 = FASTA("Name_B", "Meta_B", "C-N")
    initial = MSA([fa_1, fa_2])
    fa_reduced_1 = FASTA("Name_A", "Meta_A", "AT")
    fa_reduced_2 = FASTA("Name_B", "Meta_B", "CN")
    expected = MSA([fa_reduced_1, fa_reduced_2])
    assert_equal(initial.reduce(), expected)


def test_msa_reduce__multiple_empty_column__all_empty_column_are_removed():
    fa_1 = FASTA("Name_A", "Meta_A", "-AnTN")
    fa_2 = FASTA("Name_B", "Meta_B", "NC-NN")
    initial = MSA([fa_1, fa_2])
    fa_reduced_1 = FASTA("Name_A", "Meta_A", "AT")
    fa_reduced_2 = FASTA("Name_B", "Meta_B", "CN")
    expected = MSA([fa_reduced_1, fa_reduced_2])
    assert_equal(initial.reduce(), expected)


def test_msa_reduce__only_empty_column__none_is_returned():
    fa_1 = FASTA("Name_A", "Meta_A", "---Nn")
    fa_2 = FASTA("Name_B", "Meta_B", "Nn--N")
    initial = MSA([fa_1, fa_2])
    assert_equal(initial.reduce(), None)


###############################################################################
###############################################################################
# Tests for 'filter_singletons'

_FILTER_MSA_1 = MSA((FASTA("Seq1", "Meta1", "ACGNTYCSTG"),
                     FASTA("Seq2", "Meta2", "ACTA-WCCTG"),
                     FASTA("Seq3", "Meta3", "NCGGTYCGTC")))


def test_msa_filter_singletons__filter_by_second():
    expected = MSA((FASTA("Seq1", "Meta1", "ACnNntCcTG"),
                    FASTA("Seq2", "Meta2", "ACTA-WCCTG"),
                    FASTA("Seq3", "Meta3", "NCGGTYCGTC")))
    result = _FILTER_MSA_1.filter_singletons("Seq1", ["Seq2"])
    assert_equal(result, expected)


def test_msa_filter_singletons__filter_by_third():
    expected = MSA((FASTA("Seq1", "Meta1", "nCGNTYCgTn"),
                    FASTA("Seq2", "Meta2", "ACTA-WCCTG"),
                    FASTA("Seq3", "Meta3", "NCGGTYCGTC")))
    result = _FILTER_MSA_1.filter_singletons("Seq1", ["Seq3"])
    assert_equal(result, expected)


def test_msa_filter_singletons__filter_by_both():
    result = _FILTER_MSA_1.filter_singletons("Seq1", ["Seq2", "Seq3"])
    assert_equal(result, _FILTER_MSA_1)


def test_msa_filter_singletons__filter_by_itself():
    assert_raises(MSAError, _FILTER_MSA_1.filter_singletons, "Seq1", ["Seq1", "Seq2"])


def test_msa_filter_singletons__filter_by_nothing():
    assert_raises(ValueError, _FILTER_MSA_1.filter_singletons, "Seq1", [])


###############################################################################
###############################################################################
# Tests for 'MSA.join'

_JOIN_MSA_1 = MSA((FASTA("nc",    None, "ACG"),
                   FASTA("nm",    None, "TGA"),
                   FASTA("miRNA", None, "UCA")))
_JOIN_MSA_2 = MSA((FASTA("nc",    None, "TGA"),
                   FASTA("nm",    None, "CTT"),
                   FASTA("miRNA", None, "GAC")))
_JOIN_MSA_3 = MSA((FASTA("nc",    None, "AAG"),
                   FASTA("nm",    None, "GAG"),
                   FASTA("miRNA", None, "CAU")))


def test_msa_join__single_msa():
    result = MSA.join(_JOIN_MSA_1)
    assert_equal(result, _JOIN_MSA_1)


def test_msa_join__two_msa():
    expected = MSA((FASTA("nc",    None, "ACGTGA"),
                    FASTA("nm",    None, "TGACTT"),
                    FASTA("miRNA", None, "UCAGAC")))
    result = MSA.join(_JOIN_MSA_1, _JOIN_MSA_2)
    assert_equal(result, expected)


def test_msa_join__three_msa():
    expected = MSA((FASTA("nc",    None, "ACGTGAAAG"),
                    FASTA("nm",    None, "TGACTTGAG"),
                    FASTA("miRNA", None, "UCAGACCAU")))
    result = MSA.join(_JOIN_MSA_1, _JOIN_MSA_2, _JOIN_MSA_3)
    assert_equal(result, expected)


@nose.tools.raises(TypeError)
def test_msa_join__missing_arguments():
    MSA.join()


###############################################################################
###############################################################################
# Tests for 'MSA.from_lines'

def test_msa_from_lines__single_entry():
    lines = [">seq1", "ACG"]
    result = MSA([FASTA("seq1", None, "ACG")])
    assert_equal(MSA.from_lines(lines), result)


def test_msa_from_lines__single_entry_with_meta():
    lines = [">seq1 Meta info", "ACG"]
    expected = MSA([FASTA("seq1", "Meta info", "ACG")])
    result = MSA.from_lines(lines)
    assert_equal(result, expected)


def test_msa_from_lines__two_entries():
    lines = [">seq1", "ACG", ">seq2", "TGA"]
    expected = MSA([FASTA("seq1", None, "ACG"),
                    FASTA("seq2", None, "TGA")])
    result = MSA.from_lines(lines)
    assert_equal(result, expected)


def test_msa_from_lines__two_entries_with_meta():
    lines = [">seq1", "ACG", ">seq2 Second meta", "TGA"]
    expected = MSA([FASTA("seq1", None, "ACG"),
                    FASTA("seq2", "Second meta", "TGA")])
    result = MSA.from_lines(lines)
    assert_equal(result, expected)


@nose.tools.raises(MSAError)
def test_msa_from_lines__duplicate_names():
    MSA.from_lines([">seq1", "ACG", ">seq1", "TGA"])


@nose.tools.raises(MSAError)
def test_msa_from_lines__mismatched_lengths():
    MSA.from_lines([">seq1", "ACG", ">seq2", "TGAN"])


@nose.tools.raises(FASTAError)
def test_msa_from_lines__empty_name():
    MSA.from_lines([">", "ACG", ">seq1", "TGAN"])


###############################################################################
###############################################################################
# Tests for 'MSA.from_file'

def test_msa_from_file__uncompressed():
    expected = MSA([FASTA("This_is_FASTA!", None, "ACGTN"),
                    FASTA("This_is_ALSO_FASTA!", None, "CGTNA")])
    results = MSA.from_file(test_file("fasta_file.fasta"))
    assert_equal(results, expected)


def test_msa_from_file__compressed_gz():
    expected = MSA([FASTA("This_is_GZipped_FASTA!", None, "ACGTN"),
                    FASTA("This_is_ALSO_GZipped_FASTA!", None, "CGTNA")])
    results = MSA.from_file(test_file("fasta_file.fasta.gz"))
    assert_equal(results, expected)


def test_msa_from_file__compressed_bz2():
    expected = MSA([FASTA("This_is_BZ_FASTA!", None, "CGTNA"),
                    FASTA("This_is_ALSO_BZ_FASTA!", None,  "ACGTN")])
    results = MSA.from_file(test_file("fasta_file.fasta.bz2"))
    assert_equal(results, expected)


###############################################################################
###############################################################################
# Tests for 'MSA.split'

def test_msa_split_msa__single_group():
    msa = MSA([FASTA("seq1", None, "ACGCAT"),
               FASTA("seq2", None, "GAGTGA")])
    expected = {'1': copy.copy(msa)}
    assert_equal(msa.split("111"), expected)


def test_msa_split_msa__two_groups():
    msa = MSA([FASTA("seq1", None, "ACGCAT"),
               FASTA("seq2", None, "GAGTGA")])
    expected = {"1": MSA([FASTA("seq1", None, "ACCA"),
                          FASTA("seq2", None, "GATG")]),
                "2": MSA([FASTA("seq1", None, "GT"),
                          FASTA("seq2", None, "GA")])}
    assert_equal(msa.split("112"), expected)


def test_msa_split__three_groups():
    msa = MSA([FASTA("seq1", None, "ACGCAT"),
               FASTA("seq2", None, "GAGTGA")])
    expected = {"1": MSA([FASTA("seq1", None, "AC"),
                          FASTA("seq2", None, "GT")]),
                "2": MSA([FASTA("seq1", None, "CA"),
                          FASTA("seq2", None, "AG")]),
                "3": MSA([FASTA("seq1", None, "GT"),
                          FASTA("seq2", None, "GA")])}
    assert_equal(msa.split("123"), expected)


def test_msa_split__empty_group():
    msa = MSA([FASTA("seq1", None, "AC"),
               FASTA("seq2", None, "GA")])
    expected = {"1": MSA([FASTA("seq1", None, "A"),
                          FASTA("seq2", None, "G")]),
                "2": MSA([FASTA("seq1", None, "C"),
                          FASTA("seq2", None, "A")]),
                "3": MSA([FASTA("seq1", None, ""),
                          FASTA("seq2", None, "")])}
    assert_equal(msa.split("123"), expected)


def test_msa_split__partial_group():
    msa = MSA([FASTA("seq1", None, "ACGCA"),
               FASTA("seq2", None, "GAGTG")])
    expected = {"1": MSA([FASTA("seq1", None, "AC"),
                          FASTA("seq2", None, "GT")]),
                "2": MSA([FASTA("seq1", None, "CA"),
                          FASTA("seq2", None, "AG")]),
                "3": MSA([FASTA("seq1", None, "G"),
                          FASTA("seq2", None, "G")])}
    assert_equal(msa.split("123"), expected)


@nose.tools.raises(TypeError)
def test_msa_split_msa__no_split_by():
    msa = MSA([FASTA("seq1", None, "ACG"),
               FASTA("seq2", None, "GAT")])
    msa.split(split_by="")


###############################################################################
###############################################################################
# Tests for 'MSA.to_file'

def test_msa_to_file__complete_line_test():
    msa = MSA([FASTA("barfoo", None, "ACGATA" * 10 + "CGATAG" * 5),
               FASTA("foobar", None, "CGAATG" * 10 + "TGTCAT" * 5)])
    expected = ">barfoo\n%s\n%s\n" % ("ACGATA" * 10, "CGATAG" * 5)
    expected += ">foobar\n%s\n%s\n" % ("CGAATG" * 10, "TGTCAT" * 5)
    stringf = StringIO.StringIO()
    MSA.to_file(msa, stringf)
    assert_equal(stringf.getvalue(), expected)


###############################################################################
###############################################################################
# Tests for 'MSA.validate'

def test_msa_validate__missing_names_first():
    msa_1 = MSA(list(_JOIN_MSA_1)[:-1])
    msa_2 = copy.copy(_JOIN_MSA_2)
    assert_raises(MSAError, MSA.validate, msa_1, msa_2)


def test_msa_validate__missing_names_second():
    msa_1 = copy.copy(_JOIN_MSA_1)
    msa_2 = MSA(list(_JOIN_MSA_2)[:-1])
    assert_raises(MSAError, MSA.validate, msa_1, msa_2)


###############################################################################
###############################################################################
# Tests for 'MSA.names'

def test_msa_names():
    assert_equal(_JOIN_MSA_1.names(), set(("nc", "nm", "miRNA")))


###############################################################################
###############################################################################
# Tests for str/repr

def test_msa_repr():
    msa = MSA((FASTA("nc",    None, "ACGTA"),
               FASTA("nm",    "META", "TGAGT"),
               FASTA("miRNA", None, "UCAGA")))
    expected = "MSA(FASTA('miRNA', None, 'UCAGA'), FASTA('nc', None, 'ACGTA'), FASTA('nm', 'META', 'TGAGT'))"
    assert_equal(str(msa), expected)


def test_msa_repr__same_as_str():
    assert_equal(str(_JOIN_MSA_1), repr(_JOIN_MSA_1))



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

__version_info__ = (1, 2, 13, 7)
__version__ = '%i.%i.%i.%i' % __version_info__


def run(command=None):
    """Main entry-point for setuptools"""
    import sys
    import paleomix.main

    argv = []
    if command is not None:
        argv.append(command)
    argv.extend(sys.argv[1:])

    return paleomix.main.main(argv)


def run_bam_pipeline():
    """Legacy entry-point for setuptools"""
    return run("bam_pipeline")


def run_gtf_to_bed():
    """Legacy entry-point for setuptools"""
    return run("gtf_to_bed")


def run_phylo_pipeline():
    """Legacy entry-point for setuptools"""
    return run("phylo_pipeline")


def run_rmdup_collapsed():
    """Legacy entry-point for setuptools"""
    return run("rmdup_collapsed")


def run_trim_pipeline():
    """Legacy entry-point for setuptools"""
    return run("trim_pipeline")



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import time
import collections

import paleomix.atomiccmd.pprint as atomicpp

from paleomix.atomiccmd.command import AtomicCmd, CmdError
from paleomix.common.utilities import safe_coerce_to_tuple
from paleomix.common.fileutils import try_remove


class _CommandSet:
    def __init__(self, commands):
        self._commands = safe_coerce_to_tuple(commands)
        if not self._commands:
            raise CmdError("Empty list passed to command set")

        self._validate_commands()

    def commit(self, temp):
        committed_files = set()
        try:
            for command in self._commands:
                command.commit(temp)
                committed_files.update(command.output_files)
        except:
            # Cleanup after failed commit
            for fpath in committed_files:
                try_remove(fpath)
            raise

    def _collect_properties(key):  # pylint: disable=E0213
        def _collector(self):
            values = set()
            for command in self._commands:  # pylint: disable=W0212
                values.update(getattr(command, key))
            return values
        return property(_collector)

    input_files     = _collect_properties("input_files")
    output_files    = _collect_properties("output_files")
    auxiliary_files = _collect_properties("auxiliary_files")
    executables     = _collect_properties("executables")
    requirements    = _collect_properties("requirements")
    expected_temp_files = _collect_properties("expected_temp_files")
    optional_temp_files = _collect_properties("optional_temp_files")

    @property
    def stdout(self):
        raise CmdError("%s does not implement property 'stdout'!" \
                       % (self.__class__.__name__,))

    def terminate(self):
        for command in self._commands:
            command.terminate()

    def __str__(self):
        return atomicpp.pformat(self)

    def _validate_commands(self):
        if len(self._commands) != len(set(self._commands)):
            raise ValueError("Same command included multiple times in %s" \
                             % (self.__class__.__name__,))

        filenames = collections.defaultdict(int)
        for command in self._commands:
            for filename in command.expected_temp_files:
                filenames[filename] += 1
            for filename in command.optional_temp_files:
                filenames[filename] += 1

        clobbered = [filename for (filename, count) in filenames.items() if (count > 1)]
        if any(clobbered):
            raise CmdError("Commands clobber each others' files: %s" % (", ".join(clobbered),))


class ParallelCmds(_CommandSet):
    """This class wraps a set of AtomicCmds, running them in parallel.
    This corresponds to a set of piped commands, which only terminate
    when all parts of the pipe have terminated. For example:
    $ dmesg | grep -i segfault | gzip > log.txt.gz

    In case of any one sub-command failing, the remaining commands are
    automatically terminated. This is done to ensure that commands waiting
    on pipes are not left running indefinetly.

    Note that only AtomicCmds and ParallelCmds are allowed as
    sub-commands for this class, since the model requires non-
    blocking commands."""

    def __init__(self, commands):
        self._joinable = False

        commands = safe_coerce_to_tuple(commands)
        for command in commands:
            if not isinstance(command, (AtomicCmd, ParallelCmds)):
                raise CmdError("ParallelCmds must only contain AtomicCmds or other ParallelCmds!")
        _CommandSet.__init__(self, commands)

    def run(self, temp):
        for command in self._commands:
            command.run(temp)
        self._joinable = True

    def ready(self):
        return all(cmd.ready() for cmd in self._commands)

    def join(self):
        sleep_time = 0.05
        commands = list(enumerate(self._commands))
        return_codes = [[None]] * len(commands)
        while commands and self._joinable:
            for (index, command) in list(commands):
                if command.ready():
                    return_codes[index] = command.join()
                    commands.remove((index, command))
                    sleep_time = 0.05
                elif any(any(codes) for codes in return_codes):
                    command.terminate()
                    return_codes[index] = command.join()
                    commands.remove((index, command))
                    sleep_time = 0.05

            time.sleep(sleep_time)
            sleep_time = min(1, sleep_time * 2)
        return sum(return_codes, [])




class SequentialCmds(_CommandSet):
    """This class wraps a set of AtomicCmds, running them sequentially.
    This class therefore corresponds a set of lines in a bash script,
    each of which invokes a forground job. For example:
    $ bcftools view snps.bcf | bgzip > snps.vcf.bgz
    $ tabix snps.vcf.bgz

    The list of commands may include any type of command. Note that
    the run function only returns once each sub-command has completed.
    A command is only executed if the previous command in the sequence
    was succesfully completed, and as a consequence the return codes
    of a failed SequentialCommand may contain None."""

    def __init__(self, commands):
        self._ready = False

        commands = safe_coerce_to_tuple(commands)
        for command in commands:
            if not isinstance(command, (AtomicCmd, _CommandSet)):
                raise CmdError("ParallelCmds must only contain AtomicCmds or other ParallelCmds!")
        _CommandSet.__init__(self, commands)

    def run(self, temp):
        self._ready = False
        for command in self._commands:
            command.run(temp)
            if any(command.join()):
                break

        self._ready = True

    def ready(self):
        return self._ready

    def join(self):
        return_codes = []
        for command in self._commands:
            return_codes.extend(command.join())

        return return_codes



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from paleomix.nodes.picard import \
    ValidateBAMNode
from paleomix.nodes.samtools import \
    BAMIndexNode


def index_and_validate_bam(config, prefix, node, log_file=None,
                           create_index=True):
    input_file, index_file = _get_input_files(node, prefix['IndexFormat'])
    if not index_file and create_index:
        node = BAMIndexNode(infile=input_file,
                            index_format=prefix['IndexFormat'],
                            dependencies=node)
        index_file, = node.output_files

    ignored_checks = [
        # Ignored since we may filter out misses and low-quality hits during
        # mapping, which leads to a large proportion of missing PE mates.
        "MATE_NOT_FOUND",
        # Ignored due to high rate of false positives for lanes with few hits,
        # where high-quality reads may cause mis-identification of qualities
        "INVALID_QUALITY_FORMAT"]

    return ValidateBAMNode(config=config,
                           input_bam=input_file,
                           input_index=index_file,
                           ignored_checks=ignored_checks,
                           big_genome_mode=prefix["IndexFormat"] == ".csi",
                           output_log=log_file,
                           dependencies=node)


def _get_input_files(node, index_format):
    index_filename = None
    input_filename = None
    for filename in node.output_files:
        if filename.lower().endswith(index_format):
            index_filename = True
        elif filename.lower().endswith(".bam"):
            input_filename = filename

    return input_filename, index_filename



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import copy
import types

import paleomix.common.fileutils as fileutils
import paleomix.common.text as text


def _strand_type(value):
    if value not in ("+", "-"):
        raise ValueError("Strand must be '+' or '-', not %r" % (value,))
    return value


_BED_DEFAULTS = ("", 0, 0, "", 0, "+")
_BED_KEYS = ("contig", "start", "end", "name", "score", "strand")
_BED_TYPES = (str, int, int, str, int, _strand_type)


class BEDError(RuntimeError):
    pass


class BEDRecord(object):
    """Class for parsing and representing a BED records.

    The class has the following properties:
       .contig -> str
       .start -> int (0-based)
       .end -> int (1-based)
       .name -> str
       .score -> int
       .strand -> '+' or '-'

    These fields can also be accessed using the square brackets notation, which
    also gives access to any additional values, after the strand column. Fields
    default to 0 or empty string, except the strand which defaults to '+'.
    """

    __slots__ = ['_fields']

    def __init__(self, line=None, _len=None):
        """Constructs a BED record from a line of text. The length of the
        object matches the number of columns in the input line; in the case
        incompatible values, a BEDError exception is raised.

        The len parameter is unused, and included only for compatibility with
        pysam parser objects, such as 'asBED'. No minimum number of columns are
        required, and it is possible to construct an empty bed record.
        """
        self._fields = []

        if line:
            line = line.rstrip("\r\n").split('\t')
            for column, (value, func) in enumerate(zip(line, _BED_TYPES)):
                try:
                    self._fields.append(func(value))
                except ValueError:
                    raise BEDError("Error parsing column %i in BED record "
                                   "(%r); expected type %s, but found %r."
                                   % (column, "\t".join(line),
                                      func.__name__, value,))

            if len(line) > len(self._fields):
                self._fields.extend(line[len(self._fields):])

    def freeze(self):
        record = BEDRecord()
        record._fields = tuple(self._fields)
        return record

    def __copy__(self):
        """Needed for copy.copy to work correctly as expected."""
        record = BEDRecord()
        record._fields = copy.copy(self._fields)
        return record

    def __len__(self):
        """Returns the number of fields in the record; 0 .. N."""
        return len(self._fields)

    def __str__(self):
        """Returns a string suitable for writing to a .bed file."""
        return "\t".join(str(value) for value in self._fields)

    def __repr__(self):
        """Returns a printable representation of the record."""
        fields = []
        for name, value in zip(_BED_KEYS, self._fields):
            fields.append("%s=%r" % (name, value))

        fields.extend(repr(value) for value in self._fields[len(_BED_KEYS):])

        return "BEDRecord(%s)" % (", ".join(fields))

    def __getitem__(self, index):
        return self._fields[index]

    def __setitem__(self, index, value):
        if len(self._fields) <= index:
            defaults = _BED_DEFAULTS[len(self._fields):index + 1]
            self._fields.extend(defaults)
            while len(self._fields) <= index:
                self._fields.append("")

        if index < len(_BED_TYPES):
            if type(_BED_TYPES[index]) is type:
                if not isinstance(value, _BED_TYPES[index]):
                    raise ValueError("Expected %s for BED field %i, got %r"
                                     % (_BED_TYPES[index].__name__,
                                        index + 1, value))
            else:
                value = _BED_TYPES[index](value)

        self._fields[index] = value

    def __cmp__(self, other):
        if not isinstance(other, BEDRecord):
            return cmp(self.__class__, other.__class__)

        return cmp(self._fields, other._fields)

    @classmethod
    def _set_properties(cls):
        for index, name in enumerate(_BED_KEYS):
            setattr(cls, name, cls._new_attr(index))

    @classmethod
    def _new_attr(cls, index):
        """Returns an getter / setter property for the given value."""
        def _get(self):
            return self._fields[index]

        def _set(self, value):
            self[index] = value

        return property(_get, _set)


# Fill out properties for BEDRecord
BEDRecord._set_properties()


def read_bed_file(filename, min_columns=3, contigs=None):
    """Parses a (gzip/bzip2 compressed) BED file, and yields a sequence of
    records. Comments and empty lines are skipped. If the number of columns in
    the bed record is less than the specified ('min_columns'), a BEDError is
    raised. If a dictionary of {contig: length} is supplied, and min_columns
    is at least 6, then the coordinates are validated against the known contig
    lengths.
    """
    if min_columns < 3:
        raise ValueError("'min_columns' must be >= 3 in 'read_bed_file'")

    infinite = float("inf")
    handle = None
    try:
        handle = fileutils.open_ro(filename)

        for (line_num, line) in enumerate(handle):
            line = line.strip()
            if not line or line.startswith("#"):
                continue

            try:
                bed = BEDRecord(line)
            except ValueError, error:
                raise BEDError("Error parsing line %i in regions file:\n"
                               "  Path = %r\n  Line = %r\n\n%s"
                               % (line_num + 1, filename, line, error))

            if len(bed) < min_columns:
                url = "http://genome.ucsc.edu/FAQ/FAQformat.html#format1"
                name = repr(bed.name) if len(bed) > 3 else "unnamed record"
                raise BEDError("Region at line #%i (%s) does not "
                               "contain the expected number of fields; "
                               "the first %i fields are required. C.f. "
                               "defination at\n   %s\n\nPath = %r"
                               % (line_num, name, min_columns,
                                  url, filename))

            if contigs is None:
                contig_len = infinite
            else:
                contig_len = contigs.get(bed.contig)

            if contig_len is None:
                raise BEDError("Regions file contains contig not found "
                               "in reference:\n  Path = %r\n  Contig = "
                               "%r\n\nPlease ensure that all contig "
                               "names match the reference names!"
                               % (filename, bed.contig))
            elif not (0 <= bed.start < bed.end <= contig_len):
                raise BEDError("Regions file contains invalid region:\n"
                               "  Path   = %r\n  Contig = %r\n"
                               "  Start  = %s\n  End    = %s\n\n"
                               "Expected 0 <= Start < End <= %i!"
                               % (filename, bed.contig, bed.start,
                                  bed.end, contig_len))

            yield bed
    finally:
        if handle:
            handle.close()


def sort_bed_by_bamfile(bamfile, regions):
    """Orders a set of BED regions, such that processing matches
    (as far as possible) the layout of the BAM file. This may be
    used to ensure that extraction of regions occurs (close to)
    linearly."""
    if not regions:
        return

    references = bamfile.references
    indices = dict(zip(references, xrange(len(references))))

    def _by_bam_layout(region):
        return (indices[region.contig], region.start, region.end)
    regions.sort(key=_by_bam_layout)



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os
import StringIO

import nose.tools
from nose.tools import \
    assert_is, \
    assert_equal, \
    assert_raises, \
    assert_not_equal, \
    assert_less, \
    assert_less_equal, \
    assert_greater, \
    assert_greater_equal


from paleomix.common.testing import assert_list_equal
from paleomix.common.formats.fasta import \
     FASTA, \
     FASTAError


###############################################################################
###############################################################################

_SEQ_FRAG = "AAGTCC"  # len() = 6


def test_dir():
    return os.path.dirname(os.path.dirname(os.path.dirname(__file__)))


def test_file(*args):
    return os.path.join(test_dir(), "data", *args)


###############################################################################
###############################################################################
# Tests for FASTA constructor

def _simple_fasta_record():
    return FASTA("Dummy", "Meta-inf", "ACGT")


def test_fasta__constructor__name():
    record = _simple_fasta_record()
    assert_equal(record.name, "Dummy")


def test_fasta__constructor__meta():
    record = _simple_fasta_record()
    assert_equal(record.meta, "Meta-inf")


def test_fasta__constructor__sequence():
    record = _simple_fasta_record()
    assert_equal(record.sequence, "ACGT")


def test_fasta__constructor__name_must_be_non_empty():
    assert_raises(FASTAError, FASTA, "", None, "ACGT")


def test_fasta__constructor__name_must_be_string_type():
    assert_raises(FASTAError, FASTA, 1, None, "ACGT")


def test_fasta__constructor__name_must_be_string_type_or_none():
    assert_raises(FASTAError, FASTA, "Seq1", 1, "ACGT")


def test_fasta__constructor__sequence_must_be_string_type():
    assert_raises(FASTAError, FASTA, "Seq1", None, 1)


###############################################################################
###############################################################################
# Tests for __repr__

def test_fasta__repr__partial_line_test():
    expected = ">foobar\n%s\n" % (_SEQ_FRAG, )
    result = repr(FASTA("foobar", None, _SEQ_FRAG))
    assert_equal(result, expected)


def test_fasta__repr__complete_line_test():
    expected = ">barfoo\n%s\n" % (_SEQ_FRAG * 10, )
    result = repr(FASTA("barfoo", None, _SEQ_FRAG * 10))
    assert_equal(result, expected)


def test_fasta__repr__multiple_lines():
    expected = ">foobar\n%s\n%s\n" \
        % (_SEQ_FRAG * 10, _SEQ_FRAG * 5)
    result = repr(FASTA("foobar", None, _SEQ_FRAG * 15))
    assert_equal(result, expected)


def test_fasta__repr__partial_line_test_with_meta_information():
    expected = ">foobar my Meta-Info\n%s\n" % (_SEQ_FRAG, )
    result = repr(FASTA("foobar", "my Meta-Info", _SEQ_FRAG))
    assert_equal(result, expected)


###############################################################################
###############################################################################
# Tests for print_fasta


def test_fasta__write__partial_line():
    expected = ">foobar\n%s\n" % (_SEQ_FRAG, )
    stringf = StringIO.StringIO()
    FASTA("foobar", None, _SEQ_FRAG).write(stringf)
    assert_equal(stringf.getvalue(), expected)


def test_fasta__write__complete_line_test():
    expected = ">barfoo\n%s\n" % (_SEQ_FRAG * 10, )
    stringf = StringIO.StringIO()
    FASTA("barfoo", None, _SEQ_FRAG * 10).write(stringf)
    assert_equal(stringf.getvalue(), expected)


def test_fasta__write__multiple_lines():
    expected = ">foobar\n%s\n%s\n" \
        % (_SEQ_FRAG * 10, _SEQ_FRAG * 5)
    stringf = StringIO.StringIO()
    FASTA("foobar", None, _SEQ_FRAG * 15).write(stringf)
    assert_equal(stringf.getvalue(), expected)


###############################################################################
###############################################################################
# Tests for FASTA.from_lines

def test_fasta__from_lines__no_records():
    assert_list_equal(FASTA.from_lines([]), [])


def test_fasta__from_lines_single_record():
    lines = [">single\n", "TGTTCTCCACCGTGCACAAC\n", "CCTTCATCCA\n"]
    expected = [FASTA("single", None, "TGTTCTCCACCGTGCACAACCCTTCATCCA")]
    assert_list_equal(FASTA.from_lines(lines), expected)


def test_fasta__from_lines__multiple_records():
    lines = [">first\n",  "TGTTCTCCACCGTGCACAAC\n", "CCTTCATCCA\n",
             ">Second XT:1:0\n", "GAGAGCTCAGCTAAC\n",
             ">Third\n",  "CGCTGACCAAAAACGGACAG\n", "GGCATTCGGC\n"]
    expected = [FASTA("first", None, "TGTTCTCCACCGTGCACAACCCTTCATCCA"),
                FASTA("Second", "XT:1:0", "GAGAGCTCAGCTAAC"),
                FASTA("Third", None, "CGCTGACCAAAAACGGACAGGGCATTCGGC")]
    assert_list_equal(FASTA.from_lines(lines), expected)


@nose.tools.raises(FASTAError)
def test_fasta__from_lines__empty_record_name_only__nothing_else():
    list(FASTA.from_lines([">fasta1\n"]))


@nose.tools.raises(FASTAError)
def test_fasta__from_lines__empty_record_name_only__first():
    list(FASTA.from_lines([">fasta1\n", ">fasta2\n", "AGTC\n"]))


@nose.tools.raises(FASTAError)
def test_fasta__from_lines__empty_record__middle():
    lines = [">fasta0\n", "ACGT\n", ">fasta1\n", ">fasta2\n", "AGTC\n"]
    list(FASTA.from_lines(lines))


@nose.tools.raises(FASTAError)
def test_fasta__from_lines__empty_record_last():
    lines = [">fasta1\n", "ACGT\n", ">fasta2\n"]
    list(FASTA.from_lines(lines))


@nose.tools.raises(FASTAError)
def test_fasta__from_lines__missing_name__alone():
    lines = ["ACGT\n"]
    list(FASTA.from_lines(lines))


@nose.tools.raises(FASTAError)
def test_fasta__from_lines__missing_name__with_others():
    lines = ["ACGT\n", ">Foo\n", "ACGGTA\n"]
    list(FASTA.from_lines(lines))


@nose.tools.raises(FASTAError)
def test_fasta__from_lines__empty_name__alone():
    lines = [">\n", "ACGT\n"]
    list(FASTA.from_lines(lines))


@nose.tools.raises(FASTAError)
def test_fasta__from_lines__empty_name__with_others():
    lines = [">\n", "ACGT\n", ">Foo\n", "ACGGTA\n"]
    list(FASTA.from_lines(lines))


###############################################################################
###############################################################################
# Tests for 'FASTA.from_file'

def test_fasta__from_file__uncompressed():
    expected = [FASTA("This_is_FASTA!", None, "ACGTN"),
                FASTA("This_is_ALSO_FASTA!", None, "CGTNA")]
    results = list(FASTA.from_file(test_file("fasta_file.fasta")))
    assert_equal(results, expected)


def test_fasta__from_file__compressed_gz():
    expected = [FASTA("This_is_GZipped_FASTA!", None, "ACGTN"),
                FASTA("This_is_ALSO_GZipped_FASTA!", None, "CGTNA")]
    results = list(FASTA.from_file(test_file("fasta_file.fasta.gz")))
    assert_equal(results, expected)


def test_fasta__from_file__compressed_bz2():
    expected = [FASTA("This_is_BZ_FASTA!", None, "CGTNA"),
                FASTA("This_is_ALSO_BZ_FASTA!", None, "ACGTN")]
    results = list(FASTA.from_file(test_file("fasta_file.fasta.bz2")))
    assert_equal(results, expected)


###############################################################################
###############################################################################

def test_fasta__equality():
    assert_equal(FASTA("A", "B", "C"), FASTA("A", "B", "C"))


def test_fasta__inequality():
    assert_not_equal(FASTA("A", "B", "C"), FASTA("A", "B", "D"))
    assert_not_equal(FASTA("A", "B", "C"), FASTA("A", None, "C"))
    assert_not_equal(FASTA("A", "B", "C"), FASTA("D", "B", "C"))


def test_fasta__sorting_less_equal():
    assert not FASTA("A", "B", "C") < FASTA("A", "B", "C")
    assert_less(FASTA("A", "B", "C"), FASTA("B", "B", "C"))
    assert_less(FASTA("A", "B", "C"), FASTA("A", "C", "C"))
    assert_less(FASTA("A", "B", "C"), FASTA("A", "B", "D"))
    assert_less_equal(FASTA("A", "B", "C"), FASTA("A", "B", "C"))
    assert_less_equal(FASTA("A", "B", "C"), FASTA("B", "B", "C"))
    assert_less_equal(FASTA("A", "B", "C"), FASTA("A", "C", "C"))
    assert_less_equal(FASTA("A", "B", "C"), FASTA("A", "B", "D"))


def test_fasta__sorting_greater_equal():
    assert not FASTA("A", "B", "C") > FASTA("A", "B", "C")
    assert_greater(FASTA("B", "B", "C"), FASTA("A", "B", "C"))
    assert_greater(FASTA("A", "C", "C"), FASTA("A", "B", "C"))
    assert_greater(FASTA("A", "B", "D"), FASTA("A", "B", "C"))
    assert_greater_equal(FASTA("A", "B", "C"), FASTA("A", "B", "C"))
    assert_greater_equal(FASTA("B", "B", "C"), FASTA("A", "B", "C"))
    assert_greater_equal(FASTA("A", "C", "C"), FASTA("A", "B", "C"))
    assert_greater_equal(FASTA("A", "B", "D"), FASTA("A", "B", "C"))


def test_fasta__hash():
    assert_equal(hash(FASTA("A", "B", "C")), hash(FASTA("A", "B", "C")))
    assert_not_equal(hash(FASTA("A", "B", "C")), hash(FASTA("B", "B", "C")))
    assert_not_equal(hash(FASTA("A", "B", "C")), hash(FASTA("A", "C", "C")))
    assert_not_equal(hash(FASTA("A", "B", "C")), hash(FASTA("A", "B", "D")))


def test_fasta__unimplemented_comparison():
    assert_is(NotImplemented, FASTA("A", None, "C").__eq__(10))
    assert_is(NotImplemented, FASTA("A", None, "C").__lt__(10))
    assert_is(NotImplemented, FASTA("A", None, "C").__le__(10))
    assert_is(NotImplemented, FASTA("A", None, "C").__ge__(10))
    assert_is(NotImplemented, FASTA("A", None, "C").__gt__(10))




#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from __future__ import print_function
from __future__ import with_statement

import argparse
import collections
import heapq
import itertools
import random
import sys

import pysam

import paleomix.common.sequences as sequences
import paleomix.common.text as text

from paleomix.common.bedtools import BEDRecord
from paleomix.common.formats.fasta import FASTA


# Regions are genotyped in chunks of 1 MB strings; this is done to reduce
# overhead, as storing 1 MB of chars individually in a list adds about 45 MB of
# overhead. A 100MB chromosome would therefore require ~4.5GB.
_CHUNK_SIZE = 2 ** 20


class Pileup(object):
    def __init__(self, line):
        fields = line.split("\t")

        self.position = int(fields[1]) - 1
        self.reference = fields[2]
        self.depth = int(fields[3])
        self.observed = fields[4]


class PileupRegion(object):
    def __init__(self, tabix, chrom, start, end, padding=0,
                 dist_to_indels=0):
        assert padding >= 0, "Padding must be >= 0, not %i" % (padding,)
        self._tabix = tabix
        self._chrom = chrom
        self._start = start
        self._end = end
        self._padding = padding
        self._distance_to_indels = dist_to_indels

    def __iter__(self):
        if self._distance_to_indels <= 0:
            for pileup in self._tabix.fetch(reference=self._chrom,
                                            start=self._start,
                                            end=self._end):
                yield (pileup.position, self._sample_position(pileup))

        # Note that bed.end is a past-the-end coordinate
        start = max(0, self._start - self._padding)
        end = self._end + self._padding

        region = self._tabix.fetch(reference=self._chrom,
                                   start=start,
                                   end=end)
        pileup_buffer = collections.deque()
        blacklist = []

        # Fill buffer, and detect blacklisted sites due to indels
        for _ in xrange(max(self._padding, self._distance_to_indels) * 2):
            self._add_to_buffer(region, pileup_buffer, blacklist)

        while pileup_buffer:
            position, nucleotide = pileup_buffer.popleft()
            while blacklist and blacklist[0] < position:
                heapq.heappop(blacklist)

            if not blacklist or blacklist[0] != position:
                if self._start <= position < self._end:
                    yield (position, nucleotide)

            self._add_to_buffer(region, pileup_buffer, blacklist)

    def _add_to_buffer(self, region, pileup_buffer, blacklist):
        try:
            pileup = Pileup(region.next())
            self._collect_indels(pileup, blacklist)
            pileup_buffer.append((pileup.position,
                                  self._sample_position(pileup)))
        except StopIteration:
            pass

    def _collect_indels(self, pileup, blacklist):
        previous = None
        for (index, current) in enumerate(pileup.observed):
            if previous == '^':
                previous = current
                continue

            previous = current
            if current == "+":
                # Insertions do not themselves cover any bases
                length = 0
            elif current == "-":
                len_slice = itertools.islice(pileup.observed, index + 1, None)
                digits = "".join(itertools.takewhile(str.isdigit, len_slice))
                length = int(digits)
            else:
                continue

            # Distance is defined as sites overlapping INDELs having distance
            # 0, sites adjacent to INDELS have distance 1, etc. Note that the
            # INDEL starts on the next position of the current row.
            start = pileup.position - self._distance_to_indels + 2
            end = pileup.position + self._distance_to_indels + length
            for position in xrange(start, end):
                heapq.heappush(blacklist, position)

    @classmethod
    def _sample_position(cls, pileup):
        skip = 0
        bases = []
        observed = pileup.observed.upper()
        for current in observed:
            if skip:
                skip -= 1
            elif current in ".,":
                bases.append(pileup.reference)
            elif current in "ACGT":
                bases.append(current)
            elif current == "^":
                skip = 1
            elif current not in "$*N+-1234567890":
                assert False, current

        if not bases:
            return "N"

        return random.choice(bases)


def build_region(options, genotype, bed):
    # 'fetch' raises a ValueError if the VCF does not contain any entries for
    # the specified contig, which can occur due to low coverage.
    if bed.contig in genotype.contigs:
        region = PileupRegion(genotype,
                              chrom=bed.contig,
                              start=bed.start,
                              end=bed.end,
                              padding=options.padding,
                              dist_to_indels=options.min_distance_to_indels)

        remaining_length = (bed.end - bed.start)
        offset = bed.start

        # Genotyping is done in chunks, so that these can be reduced to strings
        # and thereby reduce the memory requirements for larger regions.
        chunk = ["N"] * min(_CHUNK_SIZE, remaining_length)

        for position, nucleotide in region:
            while position >= offset + len(chunk):
                yield "".join(chunk)
                remaining_length -= len(chunk)
                offset += len(chunk)
                chunk = ["N"] * min(_CHUNK_SIZE, remaining_length)

            chunk[position - offset] = nucleotide

        while offset < bed.end:
            yield "".join(chunk)
            remaining_length -= len(chunk)
            offset += len(chunk)
            chunk = ["N"] * min(_CHUNK_SIZE, remaining_length)

        yield "".join(chunk)
    else:
        yield "N" * (bed.end - bed.start)


def build_genes(options, genotype, regions):
    def keyfunc(bed):
        return (bed.contig, bed.name, bed.start)
    regions.sort(key=keyfunc)

    for (gene, beds) in itertools.groupby(regions, lambda x: x.name):
        sequence, beds = [], tuple(beds)
        for bed in beds:
            sequence.extend(build_region(options, genotype, bed))
        sequence = "".join(sequence)

        if any((bed.strand == "-") for bed in beds):
            assert all((bed.strand == "-") for bed in beds)

            sequence = sequences.reverse_complement(sequence)

        yield (gene, sequence)


def main(argv):
    prog = "paleomix sample_pileup"
    usage = "%s [options] --genotype in.vcf --intervals in.bed > out.fasta" \
        % (prog,)

    parser = argparse.ArgumentParser(prog=prog, usage=usage)
    parser.add_argument("--genotype", help="Tabix indexed pileup file.",
                        required=True, metavar="PILEUP")
    parser.add_argument("--intervals", help="BED file.", required=True,
                        metavar="BED")
    parser.add_argument("--padding", type=int, default=10, metavar="BED",
                        help="Number of bases to expand intervals, when "
                             "filtering based on adjacent indels "
                             "[%(default)s]")
    parser.add_argument("--min-distance-to-indels", type=int, default=5,
                        help="Variants closer than this distance from indels "
                             "are filtered; set to a negative value to "
                             "disable [%(default)s].")
    args = parser.parse_args(argv)

    genotype = pysam.Tabixfile(args.genotype)
    with open(args.intervals) as bed_file:
        intervals = text.parse_lines_by_contig(bed_file, BEDRecord)

    for (_, beds) in sorted(intervals.items()):
        for (name, sequence) in build_genes(args, genotype, beds):
            FASTA(name, None, sequence).write(sys.stdout)

    return 0


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from nose.tools import \
     assert_equal, \
     assert_raises
from flexmock import \
     flexmock

import paleomix.common.sampling as sampling


###############################################################################
###############################################################################
# weighted_sampling

def test_weighted_sampling__select_by_weight():
    def _do_select_by_weight(value, expectation):
        choices = "abc"
        weights = (1, 2, 3)
        rng = flexmock(random=lambda: value)
        iterator = sampling.weighted_sampling(choices, weights, rng)
        assert_equal(iterator.next(), expectation)

    yield _do_select_by_weight, 0.00000, 'a'
    yield _do_select_by_weight, 0.16666, 'a'  # < 1/6
    yield _do_select_by_weight, 1/6.0, 'b'
    yield _do_select_by_weight, 0.49999, 'b'  # < 3/6
    yield _do_select_by_weight, 3/6.0, 'c'
    yield _do_select_by_weight, 0.99999, 'c'


def test_weighted_sampling__empty_input_raises_value_error_for_lists():
    def _do_empty_input_raises(choices, weights):
        iterator = sampling.weighted_sampling(choices, weights)
        assert_raises(ValueError, iterator.next)

    yield _do_empty_input_raises, [], []
    yield _do_empty_input_raises, [], [1, 2]
    yield _do_empty_input_raises, [1, 2], []


def test_weighted_sampling__different_length_input_raises_value_error():
    def _do_different_length_input_raises(choices, weights):
        iterator = sampling.weighted_sampling(choices, weights)
        assert_raises(ValueError, iterator.next)

    yield _do_different_length_input_raises, [0, 1], [1, 2, 3]
    yield _do_different_length_input_raises, [0, 1, 2], [1, 2]
    yield _do_different_length_input_raises, iter([0, 1]), [1, 2, 3]
    yield _do_different_length_input_raises, [0, 1], iter([1, 2, 3])
    yield _do_different_length_input_raises, iter([0, 1]), iter([1, 2, 3])


def test_weighted_sampling__negative_weight_value_error():
    choices = range(3)
    weights = [1, -2, 3]
    iterator = sampling.weighted_sampling(choices, weights)
    assert_raises(ValueError, iterator.next)


def test_weighted_sampling__zero_weight_raises_value_error():
    choices = range(3)
    weights = [1, 0, 3]
    iterator = sampling.weighted_sampling(choices, weights)
    assert_raises(ValueError, iterator.next)


def test_weighted_sampling__non_numerical_weight_raises_type_error():
    choices = range(3)
    weights = [1, "foo", 3]
    iterator = sampling.weighted_sampling(choices, weights)
    assert_raises(TypeError, iterator.next)


###############################################################################
###############################################################################
# reservoir_sampling

def test_reservoir_sampling__select_first_item():
    rng = flexmock(randint=lambda _min, _max: 1)
    values = [1, 2]
    result = sampling.reservoir_sampling(values, 1, rng)
    assert_equal(result, [1])


def test_reservoir_sampling__select_second_item():
    rng = flexmock(randint=lambda _min, _max: 0)
    values = [1, 2]
    result = sampling.reservoir_sampling(values, 1, rng)
    assert_equal(result, [2])


def test_reservoir_sampling__upsample_equals_input():
    result = sampling.reservoir_sampling(range(5), 10)
    assert_equal(result, range(5))


def test_reservoir_sampling__downsample_to_zero():
    result = sampling.reservoir_sampling(range(5), 0)
    assert_equal(result, [])


def test_reservoir_sampling__downsample_to_negative_raises_value_error():
    assert_raises(ValueError, sampling.reservoir_sampling, range(5), -1)


def test_reservoir_sampling__downsample_to_float_raises_type_error():
    assert_raises(TypeError, sampling.reservoir_sampling, range(5), 1.0)


def test_reservoir_sampling__downsample_to_non_number_raises_type_error():
    assert_raises(TypeError, sampling.reservoir_sampling, range(5), "Eh?")



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os

from paleomix.common.formats.newick import \
     Newick
from paleomix.common.utilities import \
    safe_coerce_to_tuple
from paleomix.common.fileutils import \
    describe_files, \
    move_file
from paleomix.node import \
    Node



class NewickRerootNode(Node):
    def __init__(self, tree_files, output_file, taxa = (), dependencies = ()):
        self._output_file    = output_file
        self._tree_files     = safe_coerce_to_tuple(tree_files)
        self._reroot_on_taxa = safe_coerce_to_tuple(taxa)

        reroot_on = "midpoint"
        if self._reroot_on_taxa:
            reroot_on = repr("', '".join(sorted(self._reroot_on_taxa)))

        description  = "<NewickReroot (on %s): %s>" % \
          (reroot_on, describe_files(tree_files),)

        Node.__init__(self,
                      description  = description,
                      input_files  = self._tree_files,
                      output_files = self._output_file,
                      dependencies = dependencies)


    def _run(self, _config, temp):
        lines = []
        for tree in _read_tree_files(self._tree_files):
            if self._reroot_on_taxa:
                rooted_tree = tree.reroot_on_taxa(self._reroot_on_taxa)
            else:
                rooted_tree = tree.reroot_on_midpoint()
            lines.append(str(rooted_tree))
        lines = "\n".join(lines) + "\n"

        temp_output_file = os.path.join(temp, os.path.basename(self._output_file))
        with open(temp_output_file, "w") as handle:
            handle.write(lines)

        move_file(temp_output_file, self._output_file)




class NewickSupportNode(Node):
    def __init__(self, main_tree_files, support_tree_files, output_file, dependencies = ()):
        self._output_file        = output_file
        self._main_tree_files    = safe_coerce_to_tuple(main_tree_files)
        self._support_tree_files = safe_coerce_to_tuple(support_tree_files)
        input_files = self._main_tree_files + self._support_tree_files

        description  = "<NewickSupport: %s>" % \
          (describe_files(main_tree_files),)

        Node.__init__(self,
                      description  = description,
                      input_files  = input_files,
                      output_files = output_file,
                      dependencies = dependencies)

    def _run(self, _config, temp):
        main_trees    = _read_tree_files(self._main_tree_files)
        support_trees = _read_tree_files(self._support_tree_files)

        lines = []
        for main_tree in main_trees:
            supported_tree = main_tree.add_support(support_trees)
            lines.append(str(supported_tree))
        lines = "\n".join(lines) + "\n"

        temp_output_file = os.path.join(temp, os.path.basename(self._output_file))
        with open(temp_output_file, "w") as handle:
            handle.write(lines)

        move_file(temp_output_file, self._output_file)



def _read_tree_files(filenames):
    trees = []
    for filename in filenames:
        with open(filename) as handle:
            for line in handle:
                trees.append(Newick.from_string(line))
    return trees



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os
import sys
import types

import pysam

from paleomix.common.utilities import \
    fragment, \
    split_before, \
    Immutable, \
    TotallyOrdered
from paleomix.common.fileutils import open_ro
from paleomix.common.formats._common import FormatError


class FASTAError(FormatError):
    pass


class FASTA(TotallyOrdered, Immutable):
    def __init__(self, name, meta, sequence):
        if not (name and isinstance(name, types.StringTypes)):
            raise FASTAError("FASTA name must be a non-empty string")
        elif not (isinstance(meta, types.StringTypes) or (meta is None)):
            raise FASTAError("FASTA meta must be a string, or None")
        elif not isinstance(sequence, types.StringTypes):
            raise FASTAError("FASTA sequence must be a string")

        Immutable.__init__(self,
                           name=name,
                           meta=meta,
                           sequence=sequence)

    def write(self, fileobj=sys.stdout):
        """Prints a FASTA sequence (iterable), wrapping long sequences at 60
        characters."""
        fileobj.write(repr(self))

    @classmethod
    def from_lines(cls, lines):
        """Parses FASTA sequences found in a sequence of lines, and returns
        a tuple for each FASTA record: ((name, meta-information), sequence)
        No assumptions are made about the line-lengths."""
        lines = (line.rstrip() for line in lines)
        for record in split_before(lines, lambda v: v.startswith(">")):
            name = record[0]
            if (not name.startswith(">")) or (len(name) == 1):
                raise FASTAError("Unnamed FASTA record")
            elif len(record) == 1:
                raise FASTAError("FASTA record does not contain sequence: %s"
                                 % (name[1:],))

            # Split out any meta information
            name_and_meta = name[1:].split(None, 1)
            if len(name_and_meta) < 2:
                name_and_meta.append(None)
            name, meta = name_and_meta

            yield FASTA(name=name,
                        meta=meta,
                        sequence="".join(record[1:]))

    @classmethod
    def from_file(cls, filename):
        """Reads an unindexed FASTA file, returning a sequence of
        tuples containing the name and sequence of each entry in
        the file. The FASTA file may be GZIP/BZ2 compressed."""
        fasta_file = open_ro(filename)
        try:
            for record in FASTA.from_lines(fasta_file):
                yield record
        finally:
            fasta_file.close()

    @classmethod
    def index_and_collect_contigs(cls, filename):
        """Creates an index (.fai; if it does not already exist) for a FASTA
        file using 'pysam', and returns a dictionary of {contig: length} listed
        in that file; if the .fai file can not be created, or if the FASTA file
        contains sequences with identical names, then a FASTAError is raised.
        """
        fai_filename = filename + ".fai"
        if not os.path.exists(fai_filename):
            dirname = os.path.dirname(filename) or "."

            if not os.access(dirname, os.W_OK):
                message = \
                    "FASTA index is missing, but folder is " \
                    "not writable, so it cannot be created:\n" \
                    "  Filename = %s\n\n" \
                    "Either change permissions on the folder, or move " \
                    "the FASTA file to different location." % (filename,)
                raise FASTAError(message)

            # Use pysam to index the file
            pysam.Fastafile(filename).close()

        names = set()
        contigs = []
        with open(fai_filename) as faihandle:
            for line in faihandle:
                name, length, _ = line.split(None, 2)
                if name in names:
                    raise FASTAError("Reference contains multiple identically "
                                     "named sequences:\n  Path = %r\n  Name = "
                                     "%r\nPlease ensure that sequences have "
                                     "unique names" % (filename, name))
                names.add(name)
                contigs.append((name, int(length)))

        return contigs

    def __lt__(self, other):
        if not isinstance(other, FASTA):
            return NotImplemented

        return (self.name, self.meta, self.sequence) \
            < (other.name, other.meta, other.sequence)

    def __hash__(self):
        return hash((self.name, self.meta, self.sequence))

    def __repr__(self):
        """Returns string representation of FASTA sequence, using the standard,
        FASTA file format, wrapping long sequences at 60 characters.
        """
        name = self.name
        if self.meta:
            name = "%s %s" % (name, self.meta)
        return ">%s\n%s\n" % (name, "\n".join(fragment(60, self.sequence)))



#!/usr/bin/python
#
# Copyright (c) 2013 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from paleomix.node import \
     Node
from paleomix.common.fileutils import \
     copy_file, \
     reroot_path
from paleomix.common.utilities import \
     safe_coerce_to_tuple


class CopyOutputFilesNode(Node):
    """Copies the output-files of one or more nodes to a specified folder."""

    def __init__(self, description, destination, source_nodes):
        source_nodes = safe_coerce_to_tuple(source_nodes)

        input_files  = []
        for source_node in source_nodes:
            input_files.extend(source_node.output_files)

        output_files = [reroot_path(destination, fpath) for fpath in input_files]
        self._files  = zip(input_files, output_files)

        Node.__init__(self,
                      description  = "<Copy %s output to %r>" % (description, destination),
                      input_files  = input_files,
                      output_files = output_files,
                      dependencies = source_nodes)


    def _run(self, _config, _temp):
        for (src_file, dst_file) in self._files:
            copy_file(src_file, dst_file)



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Functions relating to the CLI interface."""
import datetime
import multiprocessing
import optparse
import os
import select
import sys
import termios
import time
import tty

import paleomix.nodegraph
import paleomix.logger
import paleomix.common.text as text

from paleomix.common.console import \
    print_msg, \
    print_debug, \
    print_info, \
    print_err, \
    print_disabled, \
    print_warn


def add_optiongroup(parser, ui_default="running", color_default="on"):
    """Adds an option-group to an OptionParser object, with options
    pertaining to logging. Note that 'initialize' expects the config
    object to have these options."""
    group = optparse.OptionGroup(parser, "Progress reporting")
    group.add_option("--progress-ui", default=ui_default, type="choice",
                     choices=("running", "progress", "summary",
                              "verbose", "quiet"),
                     help="Select method for displaying the progress of the "
                          "pipeline: 'running' = Display only currently "
                          "running nodes; 'progress' = Display changes in "
                          "state; 'summary'; one-line summary only. "
                          "[Default is '%default']")
    group.add_option("--ui-colors", default=color_default,
                     choices=("on", "off", "force"),
                     help="Enable, disable, or force the use of color codes "
                          "when printing the command-line UI. Unless forced, "
                          "colors will only be printed if STDOUT is a TTY "
                          "[Default is '%default']")
    parser.add_option_group(group)


def get_ui(ui_name):
    """Returns a UI instance by name, using the choices allowed by
    the 'add_optiongroup' function. See keys in 'UI_TYPES'."""
    ui_name = ui_name.title()
    if ui_name not in UI_TYPES:
        raise ValueError("Unknown UI type %r" % (ui_name,))
    return UI_TYPES[ui_name]()


def set_ui_colors(choice):
    import paleomix.common.console as console
    choice = choice.lower()
    if choice == "on":
        console.set_color_output(console.COLORS_ON)
    elif choice == "off":
        console.set_color_output(console.COLORS_OFF)
    elif choice == "force":
        console.set_color_output(console.COLORS_FORCED)
    else:
        raise ValueError("Unknown color setting %r" % (choice,))


class CommandLine(object):
    def __init__(self):
        self._tty_settings = None

    def __enter__(self):
        assert not self._tty_settings
        # False if the pipeline is being piped somewhere
        if sys.stdin.isatty() and sys.stdout.isatty():
            # False if the process is running in the background
            if os.getpgrp() == os.tcgetpgrp(sys.stdout.fileno()):
                try:
                    # Store old settings
                    self._tty_settings = termios.tcgetattr(sys.stdin)
                    # Disable echo
                    tty.setcbreak(sys.stdin.fileno())
                except tty.error:
                    pass  # Silently ignore failures

        return self

    def __exit__(self, _type, _value, _traceback):
        if self._tty_settings:
            # Restore settings (re-enable echo)
            termios.tcsetattr(sys.stdin, termios.TCSADRAIN, self._tty_settings)

    def process_key_presses(self, nodegraph, max_threads, ui):
        if not self._tty_settings:
            return max_threads

        help_printed = False
        old_max_threads = max_threads
        while self.poll_stdin():
            character = sys.stdin.read(1)
            if character == "+":
                max_threads = min(multiprocessing.cpu_count(), max_threads + 1)
            elif character == "-":
                max_threads = max(1, max_threads - 1)
            elif character in "lL":
                print_info(file=sys.stdout)
                progress_printer = RunningUI()
                progress_printer.max_threads = max_threads
                progress_printer.start_time = ui.start_time
                progress_printer.refresh(nodegraph)
                progress_printer.flush()
            elif character in "hH":
                if help_printed:
                    continue

                help_printed = True
                print_info("""
Commands:
  Key   Function
  h     Prints this message.
  l     Lists the currently runnning nodes.
  +     Increases the maximum number of threads by one.
  -     Decreases the maximum number of threads by one; already running tasks
        are NOT terminated if the number of threads currently used exceeds the
        resulting maximum.
""", file=sys.stdout)
            else:
                continue

        if max_threads != old_max_threads:
            print_debug("Maximum number of threads changed from %i to %i."
                        % (old_max_threads, max_threads), file=sys.stdout)

        return max_threads

    @classmethod
    def poll_stdin(cls):
        return select.select([sys.stdin], [], [], 0) == ([sys.stdin], [], [])


class BaseUI(object):
    """UI base class.

    Can be initialized, but does nothing but collect stats about
    the pipeline. Subclasses should override at least one of
    (but still call the BaseUI function) the functions 'flush',
    'finalize', and/or 'state_changed'.

    In addition, the class contains the following properties:
      - states  -- List containing the observed number of states
                   for a state-value corresponding to the index
      - threads -- Est. number of threads used by running nodes.

    These properties should be treated as read-only.
    """

    def __init__(self):
        """Basic initializer; must be called in subclasses."""
        self.states = []
        self.threads = 0
        self.max_threads = 0
        self.start_time = None
        self._end_time = None
        self._updated = True

    def flush(self):
        """Called by the user of the UI to ensure that the UI to print
        the current state of the pipeline / changes to pipeline / etc.

        Returns true if node-states have changed since last update.
        """
        if self._updated:
            self._updated = False
            return True
        return False

    def finalize(self):
        """Called by the pipeline at the termination of a run. By default,
        this function prints the location of the log-file if one was created
        during the run (e.g. if there were errors), and a summary of all nodes.
        """
        runtime = (self._end_time or 0) - (self.start_time or 0)

        if self.states[self.ERROR]:
            print_err("Done; but errors were detected ...")
        else:
            print_info("Done ...")

        print_info()
        rows = [("  Number of nodes:", sum(self.states)),
                ("  Number of done nodes:", self.states[self.DONE]),
                ("  Number of runable nodes:", self.states[self.RUNABLE]),
                ("  Number of queued nodes:", self.states[self.QUEUED]),
                ("  Number of outdated nodes:", self.states[self.OUTDATED]),
                ("  Number of failed nodes:", self.states[self.ERROR]),
                ("  Pipeline runtime:", _fmt_runtime(runtime))]

        for line in text.padded_table(rows):
            print_info(line)

        print_info("\nUse --list-output-files to view status of output files.")

        logfile = paleomix.logger.get_logfile()
        if logfile:
            print_debug("Log-file located at %r" % (logfile,))

        print_info()

    def refresh(self, nodegraph):
        """Called when the nodegraph has refreshed, causing state-counts
        to be recalculated."""
        self._updated = True
        self.states, self.threads \
            = self._count_states(nodegraph, nodegraph.iterflat())

    def state_changed(self, node, old_state, new_state, _is_primary):
        """Observer function for NodeGraph; counts states of nodes."""
        self._updated = True

        self.states[old_state] -= 1
        self.states[new_state] += 1
        if old_state == self.RUNNING:
            self.threads -= node.threads
        elif new_state == self.RUNNING:
            self.threads += node.threads

        if self.start_time is None:
            self.start_time = time.time()
        self._end_time = time.time()

    @classmethod
    def _count_states(cls, nodegraph, nodes):
        """Counts the number of each state observed for a set of nodes, and
        returns these as a list, as well as the estimated number of threads
        being used by running nodes."""
        states = [0] * nodegraph.NUMBER_OF_STATES
        threads = 0

        for node in nodes:
            state = nodegraph.get_node_state(node)

            states[state] += 1
            if state == nodegraph.RUNNING:
                threads += node.threads

        return states, threads

    def _describe_state(self):
        """ TODO """
        runtime = 0
        if self.start_time is not None:
            runtime = time.time() - self.start_time

        fields = [datetime.datetime.now().strftime("%T"),
                  ' Running ', str(self.states[self.RUNNING]), ' ',
                  ('task ' if self.states[self.RUNNING] == 1 else 'tasks '),
                  'using ~%i of max %i threads; ' % (self.threads,
                                                     self.max_threads)]

        if self.states[self.OUTDATED]:
            fields.append('%i outdated, ' % (self.states[self.OUTDATED],))

        if self.states[self.ERROR]:
            fields.append('%i failed, ' % (self.states[self.ERROR],))

        fields.extend(('%i done of %i tasks' % (self.states[self.DONE],
                                                sum(self.states),),
                       ' in ', _fmt_runtime(runtime),
                       '; press \'h\' for help.'))

        return ''.join(fields)

    DONE = paleomix.nodegraph.NodeGraph.DONE
    RUNNING = paleomix.nodegraph.NodeGraph.RUNNING
    RUNABLE = paleomix.nodegraph.NodeGraph.RUNABLE
    QUEUED = paleomix.nodegraph.NodeGraph.QUEUED
    OUTDATED = paleomix.nodegraph.NodeGraph.OUTDATED
    ERROR = paleomix.nodegraph.NodeGraph.ERROR


class RunningUI(BaseUI):
    """Prints a summary, and the list of running nodes every
    time 'flush' is called."""

    def __init__(self):
        BaseUI.__init__(self)
        self._running_nodes = []

    def flush(self):
        """See BaseUI.flush."""
        if BaseUI.flush(self) and self._running_nodes:
            self._print_header()
            for node in sorted(map(str, self._running_nodes)):
                print_info("  - %s" % (node,), file=sys.stdout)
            print_info(file=sys.stdout)

    def refresh(self, nodegraph):
        """See BaseUI.refresh."""
        BaseUI.refresh(self, nodegraph)
        self._running_nodes = []
        for node in nodegraph.iterflat():
            if nodegraph.get_node_state(node) == self.RUNNING:
                self._running_nodes.append(node)

    def state_changed(self, node, old_state, new_state, is_primary):
        """See BaseUI.state_changed."""
        BaseUI.state_changed(self, node, old_state, new_state, is_primary)

        if old_state == self.RUNNING:
            self._running_nodes.remove(node)
        elif new_state == self.RUNNING:
            self._running_nodes.append(node)

    def _print_header(self):
        print_msg('\n%s' % (self._describe_state(),), file=sys.stdout)

        logfile = paleomix.logger.get_logfile()
        if logfile:
            print_debug("  Log-file located at %r" % (logfile,),
                        file=sys.stdout)


class ProgressUI(BaseUI):
    """Progress based UI: Prints nodes when they start running; they finish
    running; or when they fail running. Changes to state resulting from the
    above is not printed. Every 20th update is followed by a summary of the
    current total progress when flush is called."""

    # Print a summery of the current state very N events
    _SUMMARY_EVERY = 20

    def __init__(self):
        self._refresh_count = ProgressUI._SUMMARY_EVERY
        self._runtimes = {}
        BaseUI.__init__(self)

    def refresh(self, nodegraph):
        """See BaseUI.refresh."""
        BaseUI.refresh(self, nodegraph)
        self._print_summary()

    def state_changed(self, node, old_state, new_state, is_primary):
        """See BaseUI.state_changed."""
        BaseUI.state_changed(self, node, old_state, new_state, is_primary)
        if is_primary and (new_state in self._DESCRIPTIONS):
            self._print_state(node, new_state)

            self._refresh_count -= 1
            if new_state == self.ERROR:
                self._refresh_count = 0

    def flush(self):
        """See BaseUI.flush."""
        if self._refresh_count <= 0:
            self._refresh_count = ProgressUI._SUMMARY_EVERY
            self._print_summary()

    def _print_summary(self):
        """Prints a summary of the pipeline progress."""
        print_msg()
        print_msg(self._describe_state(), file=sys.stdout)

        logfile = paleomix.logger.get_logfile()
        if logfile:
            print_debug("Log-file located at %r" % (logfile,), file=sys.stdout)

    def _print_state(self, node, new_state):
        state_label, print_func = self._DESCRIPTIONS[new_state]
        if new_state == self.RUNNING:
            self._runtimes[node] = time.time()
        elif new_state in (self.RUNNING, self.DONE, self.ERROR):
            state_label = "%s (%s)" % (state_label, self._get_runtime(node))

        time_label = datetime.datetime.now().strftime("%T")
        print_func("%s %s: %s" % (time_label, state_label, node),
                   file=sys.stdout)

    def _get_runtime(self, node):
        current_time = time.time()
        runtime = current_time - self._runtimes.pop(node, current_time)
        return _fmt_runtime(runtime)

    _DESCRIPTIONS = {
        BaseUI.DONE: ("Finished", print_disabled),
        BaseUI.RUNNING: ("Started", print_info),
        BaseUI.ERROR: ("Failed", print_err),
    }


class SummaryUI(BaseUI):
    def __init__(self):
        self._max_len = 0
        self._new_error = False
        BaseUI.__init__(self)

    def state_changed(self, node, old_state, new_state, is_primary):
        BaseUI.state_changed(self, node, old_state, new_state, is_primary)
        self._new_error |= (new_state == self.ERROR and is_primary)

    def flush(self):
        if BaseUI.flush(self):
            description = self._describe_state()

            self._max_len = max(len(description), self._max_len)
            print_msg("\r%s" % (description.ljust(self._max_len),), end="",
                      file=sys.stdout)

            logfile = paleomix.logger.get_logfile()
            if logfile and self._new_error:
                print_debug("\nLog-file located at %r" % (logfile,),
                            file=sys.stdout)
                self._new_error = False
            sys.stdout.flush()

    def finalize(self):
        print_msg(file=sys.stdout)
        BaseUI.finalize(self)


def _fmt_runtime(runtime):
    runtime = int(round(runtime))

    if runtime >= 3600:
        fmt = "{hours}:{mins:02}:{secs:02}s"
    elif runtime >= 60:
        fmt = "{mins}:{secs:02}s"
    else:
        fmt = "{secs}s"

    return fmt.format(hours=int(runtime) // 3600,
                      mins=(int(runtime) // 60) % 60,
                      secs=(runtime % 60))


# Different types of UIs
UI_TYPES = {
    "Verbose": RunningUI,  # Deprecated
    "Quiet": RunningUI,  # Deprecated
    "Running": RunningUI,
    "Progress": ProgressUI,
    "Summary": SummaryUI,
}



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os

import paleomix.common.rtools as rtools
import paleomix.common.versions as versions

from paleomix.common.fileutils import \
    describe_files
from paleomix.common.utilities import \
    safe_coerce_to_tuple

from paleomix.node import \
    NodeError, \
    CommandNode
from paleomix.nodes.picard import \
    MultiBAMInputNode
from paleomix.atomiccmd.builder import \
    AtomicCmdBuilder, \
    use_customizable_cli_parameters, \
    create_customizable_cli_parameters


MAPDAMAGE_VERSION = versions.Requirement(call=("mapDamage", "--version"),
                                         search=r"(\d+)\.(\d+).(\d+)",
                                         checks=versions.GE(2, 0, 1))

RSCRIPT_VERSION = versions.Requirement(call=("Rscript", "--version"),
                                       search=r"(\d+)\.(\d+).(\d+)",
                                       checks=versions.GE(2, 15, 1),
                                       priority=10)


class MapDamagePlotNode(MultiBAMInputNode):
    @create_customizable_cli_parameters
    def customize(self, config, reference, input_files, output_directory,
                  title="mapDamage", dependencies=()):
        input_files = safe_coerce_to_tuple(input_files)

        command = AtomicCmdBuilder(
            ["mapDamage", "--no-stats",
             # Prevent references with many contigs from using excessive
             # amounts of memory, at the cost of per-contig statistics:
             "--merge-reference-sequences",
             "-t", title,
             "-i", "%(TEMP_IN_BAM)s",
             "-d", "%(TEMP_DIR)s",
             "-r", "%(IN_REFERENCE)s"],

            TEMP_IN_BAM=MultiBAMInputNode.PIPE_FILE,
            IN_REFERENCE=reference,
            OUT_FREQ_3p=os.path.join(output_directory, "3pGtoA_freq.txt"),
            OUT_FREQ_5p=os.path.join(output_directory, "5pCtoT_freq.txt"),
            OUT_COMP_USER=os.path.join(output_directory, "dnacomp.txt"),
            OUT_PLOT_FRAG=os.path.join(output_directory,
                                       "Fragmisincorporation_plot.pdf"),
            OUT_PLOT_LEN=os.path.join(output_directory, "Length_plot.pdf"),
            OUT_LENGTH=os.path.join(output_directory, "lgdistribution.txt"),
            OUT_MISINCORP=os.path.join(output_directory,
                                       "misincorporation.txt"),
            OUT_LOG=os.path.join(output_directory, "Runtime_log.txt"),
            TEMP_OUT_STDOUT="pipe_mapDamage.stdout",
            TEMP_OUT_STDERR="pipe_mapDamage.stderr",

            CHECK_RSCRIPT=RSCRIPT_VERSION,
            CHECK_MAPDAMAGE=MAPDAMAGE_VERSION)

        command.add_multiple_kwargs(input_files)

        return {"command": command,
                "config": config,
                "input_files": input_files,
                "dependencies": dependencies}

    @use_customizable_cli_parameters
    def __init__(self, parameters):
        description = "<mapDamage (plots): %s -> '%s'>" \
            % (describe_files(parameters.input_files),
               parameters.output_directory)
        MultiBAMInputNode.__init__(self,
                                   config=parameters.config,
                                   input_bams=parameters.input_files,
                                   command=parameters.command.finalize(),
                                   description=description,
                                   dependencies=parameters.dependencies)

    def _teardown(self, config, temp):
        # No Length_plot.pdf file is written if there are no SE reads in the
        # input_file. In that case, we write a dummy PDF to ensure that all
        # expected files exist.
        err_message = "No length distributions are available"
        with open(os.path.join(temp, "pipe_mapDamage.stderr")) as in_handle:
            if any(line.startswith(err_message) for line in in_handle):

                fpath = os.path.join(temp, "Length_plot.pdf")
                with open(fpath, "w") as out_handle:
                    out_handle.write(_DUMMY_LENGTH_PLOT_PDF)

        MultiBAMInputNode._teardown(self, config, temp)


class MapDamageModelNode(CommandNode):
    @create_customizable_cli_parameters
    def customize(self, reference, directory, dependencies=()):
        command = AtomicCmdBuilder(
            ["mapDamage", "--stats-only",
             "-r", "%(IN_REFERENCE)s",
             "-d", "%(TEMP_DIR)s"],
            IN_REFERENCE=reference,
            TEMP_OUT_FREQ_3p="3pGtoA_freq.txt",
            TEMP_OUT_FREQ_5p="5pCtoT_freq.txt",
            TEMP_OUT_COMP_USER="dnacomp.txt",
            TEMP_OUT_MISINCORP="misincorporation.txt",
            TEMP_OUT_LOG="Runtime_log.txt",
            TEMP_OUT_STDOUT="pipe_mapDamage.stdout",
            TEMP_OUT_STDERR="pipe_mapDamage.stderr",
            OUT_COMP_GENOME=os.path.join(directory, "dnacomp_genome.csv"),
            OUT_MCMC_PROBS=os.path.join(directory,
                                        "Stats_out_MCMC_correct_prob.csv"),
            OUT_MCMC_HIST=os.path.join(directory, "Stats_out_MCMC_hist.pdf"),
            OUT_MCMC_ITER=os.path.join(directory, "Stats_out_MCMC_iter.csv"),
            OUT_MCMC_ITERSUM=os.path.join(directory,
                                          "Stats_out_MCMC_iter_summ_stat.csv"),
            OUT_MCMC_POSTPRED=os.path.join(directory,
                                           "Stats_out_MCMC_post_pred.pdf"),
            OUT_MCMC_TRACE=os.path.join(directory, "Stats_out_MCMC_trace.pdf"),

            CHECK_RSCRIPT=RSCRIPT_VERSION,
            CHECK_MAPDAMAGE=MAPDAMAGE_VERSION,
            CHECK_R_INLINE=rtools.requirement("inline"),
            CHECK_R_GGPLOT2=rtools.requirement("ggplot2"),
            CHECK_R_RCPP=rtools.requirement("Rcpp"),
            CHECK_R_GAM=rtools.requirement("gam"),
            CHECK_R_RCPPGSL=rtools.requirement("RcppGSL"))

        return {"command": command,
                "dependencies": dependencies}

    @use_customizable_cli_parameters
    def __init__(self, parameters):
        self._directory = parameters.directory

        description = "<mapDamage (model): %r>" % (parameters.directory,)
        CommandNode.__init__(self,
                             command=parameters.command.finalize(),
                             description=description,
                             dependencies=parameters.dependencies)

    def _setup(self, config, temp):
        CommandNode._setup(self, config, temp)
        for fname in ("3pGtoA_freq.txt", "5pCtoT_freq.txt", "dnacomp.txt",
                      "misincorporation.txt"):
            relpath = os.path.join(self._directory, fname)
            abspath = os.path.abspath(relpath)
            os.symlink(abspath, os.path.join(temp, fname))

    def _run(self, config, temp):
        try:
            CommandNode._run(self, config, temp)
        except NodeError, error:
            err_message = "DNA damage levels are too low"
            if self._command.join() == [1]:
                fpath = os.path.join(temp, "pipe_mapDamage.stdout")
                with open(fpath) as handle:
                    for line in handle:
                        if err_message in line:
                            line = line.strip().replace("Warning:", "ERROR:")
                            error = NodeError("%s\n\n%s" % (error, line))
                            break
            raise error


class MapDamageRescaleNode(MultiBAMInputNode):
    @create_customizable_cli_parameters
    def customize(self, config, reference, input_files, output_file, directory,
                  dependencies=()):
        input_files = safe_coerce_to_tuple(input_files)

        stats_out_fname = "Stats_out_MCMC_correct_prob.csv"
        command = AtomicCmdBuilder(["mapDamage", "--rescale-only",
                                    "-i", "%(TEMP_IN_BAM)s",
                                    "-d", "%(TEMP_DIR)s",
                                    "-r", "%(IN_REFERENCE)s",
                                    "--rescale-out", "%(OUT_BAM)s"],

                                   TEMP_IN_BAM=MultiBAMInputNode.PIPE_FILE,
                                   IN_REFERENCE=reference,
                                   TEMP_OUT_LOG="Runtime_log.txt",
                                   TEMP_OUT_CSV=stats_out_fname,
                                   OUT_BAM=output_file,
                                   CHECK_VERSION=MAPDAMAGE_VERSION)

        command.add_multiple_kwargs(input_files)

        return {"command": command,
                "config": config,
                "input_files": input_files,
                "directory": directory,
                "dependencies": dependencies}

    @use_customizable_cli_parameters
    def __init__(self, parameters):
        self._directory = parameters.directory

        description = "<mapDamage (rescale): %s -> %r>" \
            % (describe_files(parameters.input_files),
               parameters.output_file)
        MultiBAMInputNode.__init__(self,
                                   config=parameters.config,
                                   input_bams=parameters.input_files,
                                   command=parameters.command.finalize(),
                                   description=description,
                                   dependencies=parameters.dependencies)

    def _setup(self, config, temp):
        MultiBAMInputNode._setup(self, config, temp)
        for fname in ("Stats_out_MCMC_correct_prob.csv", ):
            relpath = os.path.join(self._directory, fname)
            abspath = os.path.abspath(relpath)
            os.symlink(abspath, os.path.join(temp, fname))


# Minimal PDF written if Length_plot.pdf wasn't generated
_DUMMY_LENGTH_PLOT_PDF = \
    """%PDF-1.4

1 0 obj
 <</Type /Font /Subtype /Type1 /Encoding /WinAnsiEncoding /BaseFont /Courier >>
endobj

2 0 obj
 <</Parent 4 0 R /MediaBox[0 0 450 50] /Type /Page /Contents[3 0 R ] /Resources 5 0 R >>
endobj

3 0 obj
 <</Length 138 >>
stream
  BT
    /F0 18 Tf
    20 10 Td
    (Input file(s) did not contain SE reads.) Tj
    0 20 Td
    (Length_plot.pdf not generated:) Tj
  ET
endstream
endobj

4 0 obj
 <</Type /Pages /Count 1 /Kids[2 0 R ]>>
endobj

5 0 obj
 <</ProcSet[/PDF /Text] /Font <</F0 1 0 R >>
>>
endobj

6 0 obj
 <</Type /Catalog /Pages 4 0 R >>
endobj

xref
0 7
0000000000 65535 f
0000000010 00000 n
0000000106 00000 n
0000000211 00000 n
0000000400 00000 n
0000000457 00000 n
0000000521 00000 n
trailer
 <</Size 7 /Root 6 0 R >>

startxref
571
%%EOF
"""
